"""Tests for OS function calls (Path methods) via the start/resume API.

These tests verify that Path filesystem methods correctly yield OS calls
with the right function name and arguments, and that return values from
the host are properly converted and used by Sandbox code.
"""

from pathlib import PurePosixPath
from typing import Any

import pytest
from inline_snapshot import snapshot

import ouros
from ouros import StatResult

# =============================================================================
# Basic OS call yielding
# =============================================================================


def test_path_exists_yields_oscall():
    """Path.exists() yields an OS call with correct function and path."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/test.txt").exists()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.exists')
    assert result.args == snapshot((PurePosixPath('/tmp/test.txt'),))
    assert result.kwargs == snapshot({})


def test_path_stat_yields_oscall():
    """Path.stat() yields an OS call."""
    m = ouros.Sandbox('from pathlib import Path; Path("/etc/passwd").stat()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.stat')
    assert result.args == snapshot((PurePosixPath('/etc/passwd'),))


def test_path_read_text_yields_oscall():
    """Path.read_text() yields an OS call."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/hello.txt").read_text()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.read_text')
    assert result.args == snapshot((PurePosixPath('/tmp/hello.txt'),))


# =============================================================================
# Path construction and concatenation
# =============================================================================


def test_path_concatenation():
    """Path concatenation with / operator produces correct path string."""
    code = """
from pathlib import Path
base = Path('/home')
full = base / 'user' / 'documents' / 'file.txt'
full.exists()
"""
    m = ouros.Sandbox(code)
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.args == snapshot((PurePosixPath('/home/user/documents/file.txt'),))


# =============================================================================
# Resume with return values
# =============================================================================


def test_exists_resume():
    """Resuming exists() with bool returns it to Sandbox code."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/test.txt").exists()')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=True)

    assert isinstance(result, ouros.Complete)
    assert result.output is True


def test_read_text_resume():
    """Resuming read_text() with string content returns it to Sandbox code."""
    code = """
from pathlib import Path
content = Path('/tmp/hello.txt').read_text()
'Content: ' + content
"""
    m = ouros.Sandbox(code)
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value='Hello, World!')

    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot('Content: Hello, World!')


# =============================================================================
# stat() result round-trip (Python -> Sandbox -> Python)
# =============================================================================


def test_stat_resume_and_use_in_ouros():
    """Resuming stat() with file_stat() allows Sandbox to access fields."""
    code = """
from pathlib import Path
info = Path('/tmp/file.txt').stat()
(info.st_mode, info.st_size, info[6])
"""
    m = ouros.Sandbox(code)
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    assert snapshot_result.function_name == snapshot('Path.stat')

    # Resume with a file_stat result - Sandbox accesses multiple fields
    result = snapshot_result.resume(return_value=StatResult.file_stat(1024, 0o100_644, 1234567890.0))

    assert isinstance(result, ouros.Complete)
    # st_mode=0o100_644, st_size=1024, info[6]=st_size=1024
    assert result.output == snapshot((0o100_644, 1024, 1024))


def test_stat_result_returned_from_ouros():
    """stat_result returned from Sandbox is accessible in Python."""
    code = """
from pathlib import Path
Path('/tmp/file.txt').stat()
"""
    m = ouros.Sandbox(code)
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=StatResult.file_stat(2048, 0o100_755, 1700000000.0))

    assert isinstance(result, ouros.Complete)
    stat_result = result.output

    # Access attributes on the returned namedtuple
    assert stat_result.st_mode == snapshot(0o100_755)
    assert stat_result.st_size == snapshot(2048)
    assert stat_result.st_mtime == snapshot(1700000000.0)

    # Index access works too
    assert stat_result[0] == snapshot(0o100_755)  # st_mode
    assert stat_result[6] == snapshot(2048)  # st_size


def test_stat_result():
    """stat_result repr shows field names and values."""
    code = """
from pathlib import Path
Path('/tmp/file.txt').stat()
"""
    m = ouros.Sandbox(code)
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=StatResult.file_stat(512, 0o644, 0.0))

    assert isinstance(result, ouros.Complete)
    assert repr(result.output) == snapshot(
        'StatResult(st_mode=33188, st_ino=0, st_dev=0, st_nlink=1, st_uid=0, st_gid=0, st_size=512, st_atime=0.0, st_mtime=0.0, st_ctime=0.0)'
    )
    # Should be a tuple subclass
    assert len(result.output) == 10
    assert isinstance(result.output, tuple)


# =============================================================================
# Multiple OS calls in sequence
# =============================================================================


def test_multiple_path_calls():
    """Multiple Path method calls yield multiple OS calls in sequence."""
    code = """
from pathlib import Path
p = Path('/tmp/test.txt')
exists = p.exists()
is_file = p.is_file()
(exists, is_file)
"""
    m = ouros.Sandbox(code)

    # First call: exists()
    result = m.start()
    assert isinstance(result, ouros.Snapshot)
    assert result.function_name == snapshot('Path.exists')

    # Resume exists() with True
    result = result.resume(return_value=True)
    assert isinstance(result, ouros.Snapshot)
    assert result.function_name == snapshot('Path.is_file')

    # Resume is_file() with True
    result = result.resume(return_value=True)
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot((True, True))


def test_conditional_path_calls():
    """Path calls inside conditionals work correctly."""
    code = """
from pathlib import Path
p = Path('/tmp/test.txt')
if p.exists():
    content = p.read_text()
else:
    content = 'not found'
content
"""
    m = ouros.Sandbox(code)

    # First call: exists()
    result = m.start()
    assert isinstance(result, ouros.Snapshot)
    assert result.function_name == snapshot('Path.exists')

    # Resume exists() with True - should trigger read_text()
    result = result.resume(return_value=True)
    assert isinstance(result, ouros.Snapshot)
    assert result.function_name == snapshot('Path.read_text')

    # Resume read_text() with content
    result = result.resume(return_value='file contents')
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot('file contents')


# =============================================================================
# OS call vs external function distinction
# =============================================================================


def test_os_call_vs_external_function():
    """OS calls have is_os_function=True, external functions have is_os_function=False."""
    # OS call
    m1 = ouros.Sandbox('from pathlib import Path; Path("/tmp").exists()')
    result1 = m1.start()
    assert isinstance(result1, ouros.Snapshot)
    assert result1.is_os_function is True

    # External function
    m2 = ouros.Sandbox('my_func()', external_functions=['my_func'])
    result2 = m2.start()
    assert isinstance(result2, ouros.Snapshot)
    assert result2.is_os_function is False


# =============================================================================
# os in run() method
# =============================================================================


def test_os_basic():
    """os receives function name and args, return value is used."""
    calls: list[Any] = []

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> bool:
        calls.append((function_name, args))
        return True

    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/test.txt").exists()')
    result = m.run(os=os_handler)

    assert result is True
    assert calls == snapshot([('Path.exists', (PurePosixPath('/tmp/test.txt'),))])


def test_os_stat():
    """os can return stat_result for Path.stat()."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'Path.stat':
            return StatResult.file_stat(1024, 0o644, 1700000000.0)
        return None

    code = """
from pathlib import Path
info = Path('/tmp/file.txt').stat()
(info.st_mode, info.st_size)
"""
    m = ouros.Sandbox(code)
    result = m.run(os=os_handler)

    assert result == snapshot((0o100_644, 1024))


def test_os_multiple_calls():
    """os is called for each OS operation."""
    calls: list[Any] = []

    def os_handler(
        function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None
    ) -> bool | str | None:
        calls.append(function_name)
        match function_name:
            case 'Path.exists':
                return True
            case 'Path.read_text':
                return 'file contents'
            case _:
                return None

    code = """
from pathlib import Path
p = Path('/tmp/test.txt')
if p.exists():
    result = p.read_text()
else:
    result = 'not found'
result
"""
    m = ouros.Sandbox(code)
    result = m.run(os=os_handler)

    assert result == snapshot('file contents')
    assert calls == snapshot(['Path.exists', 'Path.read_text'])


def test_os_not_provided_error():
    """Error is raised when OS call is made without os."""
    import pytest

    m = ouros.Sandbox('from pathlib import Path; Path("/tmp").exists()')
    # When no external functions and no os, run() takes the fast path
    # and OS calls raise NotImplementedError inside Sandbox
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run()
    assert str(exc_info.value) == snapshot(
        "NotImplementedError: OS function 'Path.exists' not implemented with standard execution"
    )


def test_os_not_provided_error_ext_func():
    """Error is raised when OS call is made without os."""
    import pytest

    m = ouros.Sandbox('from pathlib import Path; Path("/tmp").exists()', external_functions=['x'])
    # When no external functions and no os, run() takes the fast path
    # and OS calls raise NotImplementedError inside Sandbox
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'x': int})
    assert str(exc_info.value) == snapshot("NotImplementedError: OS function 'Path.exists' not implemented")


def test_not_callable():
    """Raise NotImplementedError inside inside ouros if so os"""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/test.txt").exists()')

    with pytest.raises(TypeError, match="TypeError: 'int' object is not callable"):
        m.run(os=123)  # type: ignore


# =============================================================================
# os.getenv() tests
# =============================================================================


def test_os_getenv_yields_oscall():
    """os.getenv() yields an OS call with correct function and args."""
    m = ouros.Sandbox('import os; os.getenv("HOME")')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('os.getenv')
    assert result.args == snapshot(('HOME', None))


def test_os_getenv_with_default_yields_oscall():
    """os.getenv() with default yields an OS call with both args."""
    m = ouros.Sandbox('import os; os.getenv("MISSING", "fallback")')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('os.getenv')
    assert result.args == snapshot(('MISSING', 'fallback'))


def test_os_getenv_callback():
    """os.getenv() with os works correctly."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> str | None:
        if function_name == 'os.getenv':
            key, default = args
            env = {'HOME': '/home/user', 'USER': 'testuser'}
            return env.get(key, default)
        return None

    m = ouros.Sandbox('import os; os.getenv("HOME")')
    result = m.run(os=os_handler)
    assert result == snapshot('/home/user')


def test_os_getenv_callback_missing():
    """os.getenv() returns None for missing env var when no default."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> str | None:
        if function_name == 'os.getenv':
            key, default = args
            env: dict[str, str] = {}
            return env.get(key, default)
        return None

    m = ouros.Sandbox('import os; os.getenv("NONEXISTENT")')
    result = m.run(os=os_handler)
    assert result is None


def test_os_getenv_callback_with_default():
    """os.getenv() uses default when env var is missing."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> str | None:
        if function_name == 'os.getenv':
            key, default = args
            env: dict[str, str] = {}
            return env.get(key, default)
        return None

    m = ouros.Sandbox('import os; os.getenv("NONEXISTENT", "default_value")')
    result = m.run(os=os_handler)
    assert result == snapshot('default_value')


# =============================================================================
# os.environ tests
# =============================================================================


def test_os_environ_yields_oscall():
    """os.environ yields an OS call with correct function name."""
    m = ouros.Sandbox('import os; os.environ')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('os.environ')
    assert result.args == snapshot(())


def test_os_environ_key_access():
    """os.environ['KEY'] works correctly after getting environ dict."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'HOME': '/home/user', 'USER': 'testuser'}
        return None

    m = ouros.Sandbox("import os; os.environ['HOME']")
    result = m.run(os=os_handler)
    assert result == snapshot('/home/user')


def test_os_environ_key_missing_raises():
    """os.environ['MISSING'] raises KeyError."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {}
        return None

    m = ouros.Sandbox("import os; os.environ['MISSING']")
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(os=os_handler)
    assert str(exc_info.value) == snapshot('KeyError: MISSING')


def test_os_environ_get_method():
    """os.environ.get() works correctly."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'HOME': '/home/user'}
        return None

    m = ouros.Sandbox("import os; os.environ.get('HOME')")
    result = m.run(os=os_handler)
    assert result == snapshot('/home/user')


def test_os_environ_get_with_default():
    """os.environ.get() with default for missing key."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {}
        return None

    m = ouros.Sandbox("import os; os.environ.get('MISSING', 'default')")
    result = m.run(os=os_handler)
    assert result == snapshot('default')


def test_os_environ_len():
    """len(os.environ) returns correct count."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'A': '1', 'B': '2', 'C': '3'}
        return None

    m = ouros.Sandbox('import os; len(os.environ)')
    result = m.run(os=os_handler)
    assert result == snapshot(3)


def test_os_environ_contains():
    """'KEY' in os.environ works correctly."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'HOME': '/home/user'}
        return None

    m = ouros.Sandbox("import os; ('HOME' in os.environ, 'MISSING' in os.environ)")
    result = m.run(os=os_handler)
    assert result == snapshot((True, False))


def test_os_environ_keys():
    """os.environ.keys() returns keys."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'HOME': '/home', 'USER': 'test'}
        return None

    m = ouros.Sandbox('import os; list(os.environ.keys())')
    result = m.run(os=os_handler)
    assert set(result) == snapshot({'HOME', 'USER'})


def test_os_environ_values():
    """os.environ.values() returns values."""

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        if function_name == 'os.environ':
            return {'A': '1', 'B': '2'}
        return None

    m = ouros.Sandbox('import os; list(os.environ.values())')
    result = m.run(os=os_handler)
    assert set(result) == snapshot({'1', '2'})


# =============================================================================
# Path write operations - write_text()
# =============================================================================


def test_path_write_text_yields_oscall():
    """Path.write_text() yields an OS call with correct function, path, and content."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/output.txt").write_text("hello world")')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.write_text')
    assert result.args == snapshot((PurePosixPath('/tmp/output.txt'), 'hello world'))


def test_path_write_text_resume():
    """Resuming write_text() with byte count returns it to Sandbox code."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/output.txt").write_text("hello")')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=5)  # write_text returns number of bytes written

    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(5)


def test_path_write_text_callback():
    """Path.write_text() with os callback works correctly."""
    written_files: dict[str, str] = {}

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> int | None:
        if function_name == 'Path.write_text':
            path, content = args
            written_files[str(path)] = content
            return len(content.encode('utf-8'))
        return None

    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/test.txt").write_text("test content")')
    result = m.run(os=os_handler)

    assert result == snapshot(12)
    assert written_files == snapshot({'/tmp/test.txt': 'test content'})


# =============================================================================
# Path write operations - write_bytes()
# =============================================================================


def test_path_write_bytes_yields_oscall():
    """Path.write_bytes() yields an OS call with correct function, path, and bytes."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/data.bin").write_bytes(b"\\x00\\x01\\x02")')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.write_bytes')
    assert result.args == snapshot((PurePosixPath('/tmp/data.bin'), b'\x00\x01\x02'))


def test_path_write_bytes_resume():
    """Resuming write_bytes() with byte count returns it to Sandbox code."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/data.bin").write_bytes(b"abc")')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=3)

    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(3)


# =============================================================================
# Path write operations - mkdir()
# =============================================================================


def test_path_mkdir_yields_oscall():
    """Path.mkdir() yields an OS call with correct function and path."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/newdir").mkdir()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.mkdir')
    assert result.args == snapshot((PurePosixPath('/tmp/newdir'),))


def test_path_mkdir_with_parents_yields_oscall():
    """Path.mkdir(parents=True) yields an OS call with kwargs."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/a/b/c").mkdir(parents=True)')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.mkdir')
    assert result.args == snapshot((PurePosixPath('/tmp/a/b/c'),))
    assert result.kwargs == snapshot({'parents': True})


def test_path_mkdir_with_exist_ok_yields_oscall():
    """Path.mkdir(exist_ok=True) yields an OS call with kwargs."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/existing").mkdir(exist_ok=True)')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.mkdir')
    assert result.kwargs == snapshot({'exist_ok': True})


def test_path_mkdir_with_both_kwargs():
    """Path.mkdir(parents=True, exist_ok=True) yields an OS call with both kwargs."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/a/b").mkdir(parents=True, exist_ok=True)')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.kwargs == snapshot({'parents': True, 'exist_ok': True})


def test_path_mkdir_resume():
    """Resuming mkdir() with None returns correctly."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/newdir").mkdir()')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=None)

    assert isinstance(result, ouros.Complete)
    assert result.output is None


# =============================================================================
# Path write operations - unlink()
# =============================================================================


def test_path_unlink_yields_oscall():
    """Path.unlink() yields an OS call with correct function and path."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/to_delete.txt").unlink()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.unlink')
    assert result.args == snapshot((PurePosixPath('/tmp/to_delete.txt'),))


def test_path_unlink_resume():
    """Resuming unlink() with None returns correctly."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/file.txt").unlink()')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=None)

    assert isinstance(result, ouros.Complete)
    assert result.output is None


# =============================================================================
# Path write operations - rmdir()
# =============================================================================


def test_path_rmdir_yields_oscall():
    """Path.rmdir() yields an OS call with correct function and path."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/empty_dir").rmdir()')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.rmdir')
    assert result.args == snapshot((PurePosixPath('/tmp/empty_dir'),))


def test_path_rmdir_resume():
    """Resuming rmdir() with None returns correctly."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/dir").rmdir()')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    result = snapshot_result.resume(return_value=None)

    assert isinstance(result, ouros.Complete)
    assert result.output is None


# =============================================================================
# Path write operations - rename()
# =============================================================================


def test_path_rename_yields_oscall():
    """Path.rename() yields an OS call with source and target paths."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/old.txt").rename(Path("/tmp/new.txt"))')
    result = m.start()

    assert isinstance(result, ouros.Snapshot)
    assert result.is_os_function is True
    assert result.function_name == snapshot('Path.rename')
    assert result.args == snapshot((PurePosixPath('/tmp/old.txt'), PurePosixPath('/tmp/new.txt')))


def test_path_rename_resume():
    """Resuming rename() returns the new path."""
    m = ouros.Sandbox('from pathlib import Path; Path("/tmp/old.txt").rename(Path("/tmp/new.txt"))')
    snapshot_result = m.start()

    assert isinstance(snapshot_result, ouros.Snapshot)
    # rename() returns None (the new Path is constructed by Sandbox)
    result = snapshot_result.resume(return_value=None)

    assert isinstance(result, ouros.Complete)
    assert result.output is None


# =============================================================================
# Write operations with os callback
# =============================================================================


def test_write_operations_callback():
    """Multiple write operations work with os callback."""
    operations: list[tuple[str, tuple[Any, ...]]] = []

    def os_handler(function_name: str, args: tuple[Any, ...], kwargs: dict[str, Any] | None = None) -> Any:
        operations.append((function_name, args))
        match function_name:
            case 'Path.mkdir':
                return None
            case 'Path.write_text':
                return len(args[1].encode('utf-8'))
            case 'Path.exists':
                return True
            case 'Path.read_text':
                return 'file content'
            case _:
                return None

    code = """
from pathlib import Path
Path('/tmp/mydir').mkdir()
Path('/tmp/mydir/file.txt').write_text('hello')
Path('/tmp/mydir/file.txt').read_text()
"""
    m = ouros.Sandbox(code)
    result = m.run(os=os_handler)

    assert result == snapshot('file content')
    assert operations == snapshot(
        [
            ('Path.mkdir', (PurePosixPath('/tmp/mydir'),)),
            ('Path.write_text', (PurePosixPath('/tmp/mydir/file.txt'), 'hello')),
            ('Path.read_text', (PurePosixPath('/tmp/mydir/file.txt'),)),
        ]
    )



# === bytes.decode() ===
assert b'hello'.decode() == 'hello', 'decode default utf-8'
assert b'hello'.decode('utf-8') == 'hello', 'decode explicit utf-8'
assert b'hello'.decode('utf8') == 'hello', 'decode utf8 variant'
assert b'hello'.decode('UTF-8') == 'hello', 'decode uppercase UTF-8'
assert b''.decode() == '', 'decode empty bytes'

# Non-ASCII UTF-8
assert b'\xc3\xa9'.decode() == '\xe9', 'decode utf-8 e-acute'
assert b'\xe4\xb8\xad'.decode() == '\u4e2d', 'decode utf-8 CJK character'

# === bytes.count() ===
assert b'hello'.count(b'l') == 2, 'count single char'
assert b'hello'.count(b'll') == 1, 'count subsequence'
assert b'hello'.count(b'x') == 0, 'count not found'
assert b'aaa'.count(b'aa') == 1, 'count non-overlapping'
assert b''.count(b'x') == 0, 'count in empty bytes'
assert b'hello'.count(b'') == 6, 'count empty subsequence'

# count with start/end
assert b'abcabc'.count(b'ab', 1) == 1, 'count with start'
assert b'abcabc'.count(b'ab', 0, 3) == 1, 'count with start and end'

# === bytes.find() ===
assert b'hello'.find(b'e') == 1, 'find single char'
assert b'hello'.find(b'll') == 2, 'find subsequence'
assert b'hello'.find(b'x') == -1, 'find not found'
assert b'hello'.find(b'') == 0, 'find empty subsequence'
assert b''.find(b'x') == -1, 'find in empty bytes'

# find with start/end
assert b'hello'.find(b'l', 3) == 3, 'find with start'
assert b'hello'.find(b'l', 0, 2) == -1, 'find with end before match'

# === bytes.index() ===
assert b'hello'.index(b'e') == 1, 'index single char'
assert b'hello'.index(b'll') == 2, 'index subsequence'
assert b'hello'.index(b'') == 0, 'index empty subsequence'

# === bytes.startswith() ===
assert b'hello'.startswith(b'he'), 'startswith true'
assert not b'hello'.startswith(b'lo'), 'startswith false'
assert b'hello'.startswith(b''), 'startswith empty'
assert b''.startswith(b''), 'empty startswith empty'
assert not b''.startswith(b'x'), 'empty startswith non-empty'

# startswith with start/end
assert b'abcdef'.startswith(b'bc', 1), 'startswith with start'
assert b'abcdef'.startswith(b'bc', 1, 3), 'startswith with start and end'
assert not b'abcdef'.startswith(b'bc', 2), 'startswith with start past match'
assert not b'abcdef'.startswith(b'abc', 0, 2), 'startswith with end before match ends'

# === bytes.endswith() ===
assert b'hello'.endswith(b'lo'), 'endswith true'
assert not b'hello'.endswith(b'he'), 'endswith false'
assert b'hello'.endswith(b''), 'endswith empty'
assert b''.endswith(b''), 'empty endswith empty'
assert not b''.endswith(b'x'), 'empty endswith non-empty'

# endswith with start/end
assert b'abcdef'.endswith(b'de', 0, 5), 'endswith with end'
assert b'abcdef'.endswith(b'cd', 1, 4), 'endswith with start and end'
assert not b'abcdef'.endswith(b'de', 0, 4), 'endswith before suffix'

# === Edge case: start > end (should not panic, treat as empty slice) ===
assert b'hello'.find(b'e', 5, 2) == -1, 'find with start > end returns -1'
assert b'hello'.count(b'l', 5, 2) == 0, 'count with start > end returns 0'
assert not b'hello'.startswith(b'h', 5, 2), 'startswith with start > end is false'
assert not b'hello'.endswith(b'o', 5, 2), 'endswith with start > end is false'

# === bytes.lower() ===
assert b'HELLO'.lower() == b'hello', 'lower basic'
assert b'Hello World'.lower() == b'hello world', 'lower mixed case'
assert b'hello'.lower() == b'hello', 'lower already lowercase'
assert b''.lower() == b'', 'lower empty'
assert b'123ABC'.lower() == b'123abc', 'lower with digits'
assert b'\x80\xff'.lower() == b'\x80\xff', 'lower non-ascii unchanged'

# === bytes.upper() ===
assert b'hello'.upper() == b'HELLO', 'upper basic'
assert b'Hello World'.upper() == b'HELLO WORLD', 'upper mixed case'
assert b'HELLO'.upper() == b'HELLO', 'upper already uppercase'
assert b''.upper() == b'', 'upper empty'
assert b'123abc'.upper() == b'123ABC', 'upper with digits'

# === bytes.capitalize() ===
assert b'hello'.capitalize() == b'Hello', 'capitalize basic'
assert b'HELLO'.capitalize() == b'Hello', 'capitalize uppercase'
assert b'hELLO wORLD'.capitalize() == b'Hello world', 'capitalize mixed'
assert b''.capitalize() == b'', 'capitalize empty'
assert b'123hello'.capitalize() == b'123hello', 'capitalize starting with digit'

# === bytes.title() ===
assert b'hello world'.title() == b'Hello World', 'title basic'
assert b'HELLO WORLD'.title() == b'Hello World', 'title uppercase'
assert b"they're bill's".title() == b"They'Re Bill'S", 'title with apostrophe'
assert b''.title() == b'', 'title empty'

# === bytes.swapcase() ===
assert b'Hello World'.swapcase() == b'hELLO wORLD', 'swapcase basic'
assert b'HELLO'.swapcase() == b'hello', 'swapcase uppercase'
assert b'hello'.swapcase() == b'HELLO', 'swapcase lowercase'
assert b''.swapcase() == b'', 'swapcase empty'

# === bytes.isalpha() ===
assert b'hello'.isalpha(), 'isalpha all letters'
assert not b'hello123'.isalpha(), 'isalpha with digits'
assert not b'hello world'.isalpha(), 'isalpha with space'
assert not b''.isalpha(), 'isalpha empty is false'
assert b'ABC'.isalpha(), 'isalpha uppercase'

# === bytes.isdigit() ===
assert b'123'.isdigit(), 'isdigit all digits'
assert not b'123abc'.isdigit(), 'isdigit with letters'
assert not b''.isdigit(), 'isdigit empty is false'

# === bytes.isalnum() ===
assert b'hello123'.isalnum(), 'isalnum letters and digits'
assert b'hello'.isalnum(), 'isalnum all letters'
assert b'123'.isalnum(), 'isalnum all digits'
assert not b'hello world'.isalnum(), 'isalnum with space'
assert not b''.isalnum(), 'isalnum empty is false'

# === bytes.isspace() ===
assert b' \t\n\r'.isspace(), 'isspace whitespace chars'
assert not b'hello'.isspace(), 'isspace not all whitespace'
assert not b''.isspace(), 'isspace empty is false'
assert b' '.isspace(), 'isspace single space'

# === bytes.islower() ===
assert b'hello'.islower(), 'islower all lowercase'
assert b'hello123'.islower(), 'islower with digits'
assert not b'Hello'.islower(), 'islower with uppercase'
assert not b'HELLO'.islower(), 'islower all uppercase'
assert not b''.islower(), 'islower empty is false'
assert not b'123'.islower(), 'islower no cased chars is false'

# === bytes.isupper() ===
assert b'HELLO'.isupper(), 'isupper all uppercase'
assert b'HELLO123'.isupper(), 'isupper with digits'
assert not b'Hello'.isupper(), 'isupper with lowercase'
assert not b'hello'.isupper(), 'isupper all lowercase'
assert not b''.isupper(), 'isupper empty is false'
assert not b'123'.isupper(), 'isupper no cased chars is false'

# === bytes.isascii() ===
assert b'hello'.isascii(), 'isascii all ascii'
assert b''.isascii(), 'isascii empty is true'
assert b'\x00\x7f'.isascii(), 'isascii boundary values'
assert not b'\x80'.isascii(), 'isascii non-ascii byte'
assert not b'hello\xff'.isascii(), 'isascii with non-ascii'

# === bytes.istitle() ===
assert b'Hello World'.istitle(), 'istitle basic'
assert not b'hello world'.istitle(), 'istitle lowercase'
assert not b'HELLO WORLD'.istitle(), 'istitle uppercase'
assert b'Hello'.istitle(), 'istitle single word'
assert not b''.istitle(), 'istitle empty is false'

# === bytes.rfind() ===
assert b'hello'.rfind(b'l') == 3, 'rfind finds last occurrence'
assert b'hello'.rfind(b'x') == -1, 'rfind not found'
assert b'hello'.rfind(b'') == 5, 'rfind empty at end'
assert b'aaaa'.rfind(b'aa') == 2, 'rfind non-overlapping from right'
assert b'hello'.rfind(b'l', 0, 3) == 2, 'rfind with range'

# === bytes.rindex() ===
assert b'hello'.rindex(b'l') == 3, 'rindex finds last occurrence'
assert b'hello'.rindex(b'') == 5, 'rindex empty at end'

# === bytes.strip() ===
assert b'  hello  '.strip() == b'hello', 'strip whitespace'
assert b'hello'.strip() == b'hello', 'strip no whitespace'
assert b'xxxhelloxxx'.strip(b'x') == b'hello', 'strip custom chars'
assert b''.strip() == b'', 'strip empty'
assert b'   '.strip() == b'', 'strip all whitespace'

# === bytes.lstrip() ===
assert b'  hello  '.lstrip() == b'hello  ', 'lstrip whitespace'
assert b'xxxhello'.lstrip(b'x') == b'hello', 'lstrip custom chars'
assert b''.lstrip() == b'', 'lstrip empty'

# === bytes.rstrip() ===
assert b'  hello  '.rstrip() == b'  hello', 'rstrip whitespace'
assert b'helloxxx'.rstrip(b'x') == b'hello', 'rstrip custom chars'
assert b''.rstrip() == b'', 'rstrip empty'

# === bytes.removeprefix() ===
assert b'hello'.removeprefix(b'he') == b'llo', 'removeprefix found'
assert b'hello'.removeprefix(b'xxx') == b'hello', 'removeprefix not found'
assert b'hello'.removeprefix(b'') == b'hello', 'removeprefix empty'
assert b''.removeprefix(b'x') == b'', 'removeprefix empty bytes'

# === bytes.removesuffix() ===
assert b'hello'.removesuffix(b'lo') == b'hel', 'removesuffix found'
assert b'hello'.removesuffix(b'xxx') == b'hello', 'removesuffix not found'
assert b'hello'.removesuffix(b'') == b'hello', 'removesuffix empty'
assert b''.removesuffix(b'x') == b'', 'removesuffix empty bytes'

# === bytes.split() ===
assert b'a,b,c'.split(b',') == [b'a', b'b', b'c'], 'split basic'
assert b'a b c'.split() == [b'a', b'b', b'c'], 'split whitespace'
assert b'a  b  c'.split() == [b'a', b'b', b'c'], 'split multiple whitespace'
assert b'a,b,c'.split(b',', 1) == [b'a', b'b,c'], 'split maxsplit'
assert b''.split() == [], 'split empty bytes'
assert b'hello'.split(b'x') == [b'hello'], 'split not found'

# === bytes.rsplit() ===
assert b'a,b,c'.rsplit(b',') == [b'a', b'b', b'c'], 'rsplit basic'
assert b'a,b,c'.rsplit(b',', 1) == [b'a,b', b'c'], 'rsplit maxsplit'
assert b'a b c'.rsplit() == [b'a', b'b', b'c'], 'rsplit whitespace'

# === bytes.splitlines() ===
assert b'a\nb\nc'.splitlines() == [b'a', b'b', b'c'], 'splitlines newline'
assert b'a\r\nb\rc'.splitlines() == [b'a', b'b', b'c'], 'splitlines mixed'
assert b'a\nb\n'.splitlines() == [b'a', b'b'], 'splitlines trailing'
assert b'a\nb'.splitlines(True) == [b'a\n', b'b'], 'splitlines keepends'
assert b''.splitlines() == [], 'splitlines empty'

# === bytes.partition() ===
assert b'hello world'.partition(b' ') == (b'hello', b' ', b'world'), 'partition found'
assert b'hello'.partition(b'x') == (b'hello', b'', b''), 'partition not found'
assert b'hello world here'.partition(b' ') == (b'hello', b' ', b'world here'), 'partition first'

# === bytes.rpartition() ===
assert b'hello world'.rpartition(b' ') == (b'hello', b' ', b'world'), 'rpartition found'
assert b'hello'.rpartition(b'x') == (b'', b'', b'hello'), 'rpartition not found'
assert b'hello world here'.rpartition(b' ') == (b'hello world', b' ', b'here'), 'rpartition last'

# === bytes.replace() ===
assert b'hello'.replace(b'l', b'L') == b'heLLo', 'replace all'
assert b'hello'.replace(b'l', b'L', 1) == b'heLlo', 'replace count'
assert b'hello'.replace(b'x', b'y') == b'hello', 'replace not found'
assert b'aaa'.replace(b'a', b'bb') == b'bbbbbb', 'replace longer'
assert b'aaa'.replace(b'aa', b'b') == b'ba', 'replace non-overlapping'

# === bytes.center() ===
assert b'hello'.center(10) == b'  hello   ', 'center basic'
assert b'hello'.center(10, b'*') == b'**hello***', 'center fillbyte'
assert b'hello'.center(3) == b'hello', 'center too short'

# === bytes.ljust() ===
assert b'hello'.ljust(10) == b'hello     ', 'ljust basic'
assert b'hello'.ljust(10, b'*') == b'hello*****', 'ljust fillbyte'
assert b'hello'.ljust(3) == b'hello', 'ljust too short'

# === bytes.rjust() ===
assert b'hello'.rjust(10) == b'     hello', 'rjust basic'
assert b'hello'.rjust(10, b'*') == b'*****hello', 'rjust fillbyte'
assert b'hello'.rjust(3) == b'hello', 'rjust too short'

# === bytes.zfill() ===
assert b'42'.zfill(5) == b'00042', 'zfill basic'
assert b'-42'.zfill(5) == b'-0042', 'zfill negative'
assert b'+42'.zfill(5) == b'+0042', 'zfill positive'
assert b'hello'.zfill(3) == b'hello', 'zfill too short'

# === bytes.join() ===
assert b','.join([b'a', b'b', b'c']) == b'a,b,c', 'join list'
assert b''.join([b'a', b'b']) == b'ab', 'join empty separator'
assert b','.join([]) == b'', 'join empty iterable'
assert b'-'.join([b'hello']) == b'hello', 'join single item'

# === bytes.hex() ===
assert b'\xde\xad\xbe\xef'.hex() == 'deadbeef', 'hex basic'
assert b''.hex() == '', 'hex empty'
assert b'AB'.hex() == '4142', 'hex letters'
assert b'\x00\xff'.hex() == '00ff', 'hex boundary'
assert b'\xde\xad\xbe\xef'.hex(':') == 'de:ad:be:ef', 'hex with separator'
assert b'\xde\xad\xbe\xef'.hex(':', 2) == 'dead:beef', 'hex with bytes_per_sep'
# Test positive bytes_per_sep (partial group at start)
assert b'\x01\x02\x03\x04\x05'.hex(':', 2) == '01:0203:0405', 'hex +2 odd bytes'
assert b'\x01\x02\x03'.hex(':', 2) == '01:0203', 'hex +2 three bytes'
# Test negative bytes_per_sep (partial group at end)
assert b'\x01\x02\x03\x04\x05'.hex(':', -2) == '0102:0304:05', 'hex -2 odd bytes'
assert b'\x01\x02\x03'.hex(':', -2) == '0102:03', 'hex -2 three bytes'

# === bytes.fromhex() ===
assert bytes.fromhex('deadbeef') == b'\xde\xad\xbe\xef', 'fromhex basic'
assert bytes.fromhex('DEADBEEF') == b'\xde\xad\xbe\xef', 'fromhex uppercase'
assert bytes.fromhex('') == b'', 'fromhex empty'
assert bytes.fromhex('de ad be ef') == b'\xde\xad\xbe\xef', 'fromhex with spaces'
assert bytes.fromhex('4142') == b'AB', 'fromhex letters'

# === bytes.fromhex() with whitespace ===
# Whitespace is only allowed BETWEEN byte pairs, not within a pair
assert bytes.fromhex(' 01 ') == b'\x01', 'fromhex whitespace around bytes is stripped'
assert bytes.fromhex('01 23') == b'\x01\x23', 'fromhex whitespace between byte pairs'

# === bytes.fromhex() errors ===
# Odd number of hex digits (no invalid chars, just odd count)
try:
    bytes.fromhex('0')
    assert False, 'fromhex odd digits should error'
except ValueError as e:
    assert str(e) == 'fromhex() arg must contain an even number of hexadecimal digits', (
        f'fromhex odd digits message, error: {e}'
    )

try:
    bytes.fromhex(' 0')
    assert False, 'fromhex odd digits after whitespace should error'
except ValueError as e:
    assert str(e) == 'fromhex() arg must contain an even number of hexadecimal digits', (
        f'fromhex odd digits after whitespace message, error: {e}'
    )

# Whitespace within a byte pair is invalid (space is not a hex digit)
try:
    bytes.fromhex('0 1')
    assert False, 'fromhex whitespace within pair should error'
except ValueError as e:
    assert str(e) == 'non-hexadecimal number found in fromhex() arg at position 1', (
        f'fromhex whitespace within pair message, error: {e}'
    )

# Invalid hex character
try:
    bytes.fromhex('0g')
    assert False, 'fromhex invalid hex char should error'
except ValueError as e:
    assert str(e) == 'non-hexadecimal number found in fromhex() arg at position 1', (
        f'fromhex invalid hex char message, error: {e}'
    )

# === bytes.fromhex() instance access ===
# fromhex is a classmethod but should also work on instances
assert b''.fromhex('4142') == b'AB', 'fromhex on bytes instance'
assert b'hello'.fromhex('deadbeef') == b'\xde\xad\xbe\xef', 'fromhex on non-empty instance'

# === bytes.startswith/endswith with tuple of prefixes ===
assert b'hello'.startswith((b'he', b'wo')), 'startswith tuple first match'
assert b'hello'.startswith((b'wo', b'he')), 'startswith tuple second match'
assert not b'hello'.startswith((b'wo', b'ab')), 'startswith tuple no match'
assert b'hello'.startswith((b'',)), 'startswith tuple with empty bytes'
assert b'hello'.startswith((b'hello', b'world')), 'startswith tuple exact match'

assert b'hello'.endswith((b'lo', b'ld')), 'endswith tuple first match'
assert b'hello'.endswith((b'ld', b'lo')), 'endswith tuple second match'
assert not b'hello'.endswith((b'he', b'ab')), 'endswith tuple no match'
assert b'hello'.endswith((b'',)), 'endswith tuple with empty bytes'
assert b'hello'.endswith((b'hello', b'world')), 'endswith tuple exact match'

# startswith/endswith tuple with start/end
assert b'abcdef'.startswith((b'bc', b'cd'), 1), 'startswith tuple with start'
assert b'abcdef'.endswith((b'de', b'cd'), 0, 5), 'endswith tuple with end'

# === Empty-substring edge cases ===
# Edge case: start == len (boundary) - this works
assert b'hello'.find(b'', 5) == 5, 'find empty at len returns len'
assert b'hello'.count(b'', 5) == 1, 'count empty at len returns 1'
assert b'hello'.startswith(b'', 5), 'startswith empty at len is true'
assert b'hello'.endswith(b'', 5), 'endswith empty at len is true'

# TODO: These edge cases when start > len need to be fixed
# CPython returns -1/0/False for these, currently Ouros doesn't handle this correctly
# assert b'hello'.find(b'', 10) == -1, 'find empty when start > len returns -1'
# assert b'hello'.count(b'', 10) == 0, 'count empty when start > len returns 0'
# assert not b'hello'.startswith(b'', 10), 'startswith empty when start > len is false'
# assert not b'hello'.endswith(b'', 10), 'endswith empty when start > len is false'
# assert b'hello'.rfind(b'', 10) == -1, 'rfind empty when start > len returns -1'

# === bytes.hex() non-ASCII separator errors ===
try:
    b'\x01\x02'.hex('\xff')
    assert False, 'hex with non-ASCII separator should error'
except ValueError as e:
    # CPython uses 'sep must be ASCII.' with period
    msg = str(e)
    assert 'sep' in msg.lower() and 'ascii' in msg.lower(), f'hex non-ASCII sep message, error: {e}'

# === bytes.decode() with errors argument ===
# Valid errors values
assert b'hello'.decode('utf-8', 'strict') == 'hello', 'decode with strict errors'
assert b'hello'.decode('utf-8', 'ignore') == 'hello', 'decode with ignore errors'
assert b'hello'.decode('utf-8', 'replace') == 'hello', 'decode with replace errors'

# TODO: errors argument type validation - CPython raises TypeError for non-string errors
# This is not implemented yet
# try:
#     b'hello'.decode('utf-8', 123)
#     assert False, 'decode with non-string errors should error'
# except TypeError as e:
#     assert 'str' in str(e), f'decode errors type error should mention str, error: {e}'

# === Error message for unknown classmethod ===
# Error message should say 'bytes' not 'type'
try:
    bytes.nonexistent()
    assert False, 'should raise AttributeError'
except AttributeError as e:
    msg = str(e)
    assert 'bytes' in msg, f'error should mention bytes, got: {e}'
    assert 'nonexistent' in msg, f'error should mention method name, got: {e}'



# Security tests for class implementation
# These should trigger errors/limits, NOT crash

# === Recursion Protection ===

# Test: Infinite __init__ recursion
class InfiniteInit:
    def __init__(self):
        InfiniteInit()


try:
    InfiniteInit()
    assert False, 'should have raised RecursionError'
except RecursionError:
    pass


# Test: __init__ mutual recursion
class MutA:
    def __init__(self):
        MutB()


class MutB:
    def __init__(self):
        MutA()


try:
    MutA()
    assert False, 'should have raised RecursionError'
except RecursionError:
    pass


# Test: Circular references
class Node:
    def __init__(self):
        self.ref = None


a = Node()
b = Node()
a.ref = b
b.ref = a
# Should not crash when collected


# Test: Class with no methods
class Empty:
    pass


obj = Empty()
assert type(obj).__name__ == 'Empty'


# Test: Instance keeps class alive
def make_instance():
    class Local:
        pass

    return Local()


obj = make_instance()
assert type(obj).__name__ == 'Local'

print('All security tests passed')



# call-external
# External calls in list and dict literals

# External call in list literal
lst = [add_ints(1, 2), add_ints(3, 4)]
assert lst[0] == 3, 'ext call in list literal [0]'
assert lst[1] == 7, 'ext call in list literal [1]'

# External call in tuple literal
tup = (add_ints(1, 1), add_ints(2, 2))
assert tup[0] == 2, 'ext call in tuple literal [0]'
assert tup[1] == 4, 'ext call in tuple literal [1]'

# External call in dict value
d = {'a': add_ints(5, 5), 'b': add_ints(10, 10)}
assert d['a'] == 10, 'ext call in dict value a'
assert d['b'] == 20, 'ext call in dict value b'



','.join(['a'], ['b'])
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__join_too_many_args.py", line 1, in <module>
    ','.join(['a'], ['b'])
    ~~~~~~~~~~~~~~~~~~~~~~
TypeError: str.join() takes exactly one argument (2 given)
"""



import functools
import sys
from functools import partial, reduce

# === reduce with lambda ===
result = functools.reduce(lambda a, b: a + b, [1, 2, 3, 4])
assert result == 10, 'reduce sum'

result = functools.reduce(lambda a, b: a * b, [1, 2, 3, 4])
assert result == 24, 'reduce product'

# === reduce with initial ===
result = functools.reduce(lambda a, b: a + b, [1, 2, 3], 10)
assert result == 16, 'reduce with initial'

# === reduce single element ===
result = functools.reduce(lambda a, b: a + b, [42])
assert result == 42, 'reduce single'

# === from import reduce ===
assert reduce(lambda a, b: a + b, [1, 2]) == 3, 'from import reduce'

# === partial: basic usage ===
add = partial(lambda a, b: a + b, 10)
assert add(5) == 15, 'partial basic'

# === partial: multiple pre-applied args ===
add3 = partial(lambda a, b, c: a + b + c, 1, 2)
assert add3(3) == 6, 'partial multiple args'

# === partial: no extra args ===
greet = partial(lambda: 'hello')
assert greet() == 'hello', 'partial no extra args'

# === partial: with builtin-like lambda ===
double = partial(lambda x, factor: x * factor, factor=2)
# Note: kwargs not fully supported in partial, test positional only
mul = partial(lambda x, y: x * y, 3)
assert mul(4) == 12, 'partial multiply'

# === partial: chained ===
add10 = partial(lambda a, b: a + b, 10)
assert add10(20) == 30, 'partial chained call'

# === cmp_to_key: basic import ===
from functools import cmp_to_key

key_fn = cmp_to_key(lambda a, b: a - b)
assert key_fn is not None, 'cmp_to_key returns object'

# === functools wrapper constants ===
wrapper_assignments = functools.WRAPPER_ASSIGNMENTS
assert isinstance(wrapper_assignments, tuple), 'WRAPPER_ASSIGNMENTS is a tuple'
for required_name in ('__module__', '__name__', '__qualname__', '__doc__'):
    assert required_name in wrapper_assignments, f'WRAPPER_ASSIGNMENTS contains {required_name}'

# CPython 3.14 switched to '__annotate__'; Ouros currently uses '__annotations__'.
assert '__annotations__' in wrapper_assignments or '__annotate__' in wrapper_assignments, (
    'WRAPPER_ASSIGNMENTS contains annotation attribute'
)

# Keep a strict Ouros assertion for the compatibility target in this task.
if sys.platform == 'ouros':
    assert '__annotate__' in wrapper_assignments, 'Ouros WRAPPER_ASSIGNMENTS includes __annotate__ (PEP 749)'

assert functools.WRAPPER_UPDATES == ('__dict__',), 'WRAPPER_UPDATES tuple matches stdlib defaults'

# === lru_cache/cache wiring ===
lru_factory = functools.lru_cache()
assert lru_factory is not None, 'lru_cache() returns a decorator/wrapper object'
if sys.platform == 'ouros':
    assert repr(lru_factory) == '<functools.lru_cache object>', 'Ouros lru_cache() returns lru wrapper object'

lru_direct = functools.lru_cache(lambda x: x)
assert lru_direct is not None, 'lru_cache(func) returns a wrapped callable'
if sys.platform == 'ouros':
    assert repr(lru_direct) == '<functools.lru_cache object>', 'Ouros lru_cache(func) returns lru wrapper object'

cache_direct = functools.cache(lambda x: x)
assert cache_direct is not None, 'cache(func) returns an unbounded cached callable/wrapper'
if sys.platform == 'ouros':
    assert repr(cache_direct) == '<functools.lru_cache object>', 'Ouros cache() aliases lru_cache(maxsize=None)'

# === wraps/update_wrapper wiring ===
wrapped = lambda value: value + 1
wrapper = lambda value: value * 2

wraps_factory = functools.wraps(wrapped)
assert wraps_factory is not None, 'wraps() returns decorator factory'
if sys.platform == 'ouros':
    assert repr(wraps_factory) == '<functools.wraps object>', 'Ouros wraps() returns wraps decorator object'

updated = functools.update_wrapper(wrapper, wrapped)
assert updated.__wrapped__ is wrapped, 'update_wrapper stores __wrapped__'
assert updated.__name__ == wrapped.__name__, 'update_wrapper copies __name__'
assert updated.__module__ == wrapped.__module__, 'update_wrapper copies __module__'
assert updated.__qualname__ == wrapped.__qualname__, 'update_wrapper copies __qualname__'
assert updated.__doc__ == wrapped.__doc__, 'update_wrapper copies __doc__'


# === total_ordering wiring ===
class TotalOrderingSample:
    def __lt__(self, other):
        return False


ordered_class = functools.total_ordering(TotalOrderingSample)
assert ordered_class is TotalOrderingSample, 'total_ordering returns the decorated class'

# === Placeholder sentinel + partial substitution ===
placeholder = functools.Placeholder
with_holes = functools.partial(lambda a, b, c: (a, b, c), placeholder, 2, 3)
assert with_holes(1) == (1, 2, 3), 'partial should fill Placeholder slots from call args'


# === cached_property ===
class CachedPropertyExample:
    def __init__(self):
        self.calls = 0

    @functools.cached_property
    def value(self):
        self.calls += 1
        return self.calls * 10


cached = CachedPropertyExample()
assert cached.value == 10, 'cached_property computes value on first access'
assert cached.value == 10, 'cached_property returns cached value on repeated access'
assert cached.calls == 1, 'cached_property should execute wrapped function once'


# === singledispatch ===
@functools.singledispatch
def dispatch_value(value):
    return 'default'


@dispatch_value.register(int)
def _(value):
    return 'int'


@dispatch_value.register(str)
def _(value):
    return 'str'


def dispatch_bool(value):
    return 'bool'


dispatch_value.register(bool, dispatch_bool)
assert dispatch_value(1) == 'int', 'singledispatch should use int registration'
assert dispatch_value('x') == 'str', 'singledispatch should use str registration'
assert dispatch_value(True) == 'bool', 'singledispatch register(cls, func) should work'
assert dispatch_value(1.5) == 'default', 'singledispatch should fall back to default implementation'
assert dispatch_value.dispatch(int)(10) == 'int', 'singledispatch dispatch(type) should return implementation'


# === singledispatchmethod ===
class SingleDispatchMethodExample:
    @functools.singledispatchmethod
    def render(self, value):
        return 'default'

    @render.register(int)
    def _(self, value):
        return 'int'

    @render.register(str)
    def _(self, value):
        return 'str'


single_method = SingleDispatchMethodExample()
assert single_method.render(1) == 'int', 'singledispatchmethod should dispatch on first non-self arg'
assert single_method.render('x') == 'str', 'singledispatchmethod should dispatch registered str overload'
assert single_method.render(1.5) == 'default', 'singledispatchmethod should fall back to default implementation'


# === partialmethod ===
class PartialMethodExample:
    def add(self, a, b, c=0):
        return a + b + c

    add_five = functools.partialmethod(add, 5)
    add_five_and_six = functools.partialmethod(add, 5, 6)
    add_with_kw = functools.partialmethod(add, 5, c=7)


partial_method = PartialMethodExample()
assert partial_method.add_five(4) == 9, 'partialmethod should bind instance and pre-applied positional args'
assert partial_method.add_five_and_six(3) == 14, 'partialmethod should support multiple pre-applied args'
assert partial_method.add_with_kw(2) == 14, 'partialmethod should merge pre-applied kwargs with call args'



# === Basic f-strings ===
assert f'hello' == 'hello', 'basic f-string'
assert f'' == '', 'empty f-string'

# === Simple interpolation ===
x = 'world'
assert f'hello {x}' == 'hello world', 'simple interpolation'

# multiple interpolations
a = 1
b = 2
assert f'{a} + {b} = {a + b}' == '1 + 2 = 3', 'multiple interpolations'

# expression in f-string
assert f'{1 + 2 + 3}' == '6', 'expression'

# === Value types ===
# list value
x = [1, 2, 3]
assert f'list: {x}' == 'list: [1, 2, 3]', 'list value'

# bool value
x = True
assert f'value: {x}' == 'value: True', 'bool value'

# int value
assert f'{42}' == '42', 'int value'

# float value
assert f'{3.14}' == '3.14', 'float value'

# None value
assert f'{None}' == 'None', 'None value'

# === Conversion flags (!s, !r, !a) ===
# conversion !s (str)
assert f'{42!s}' == '42', 'conversion !s'

# conversion !r (repr)
assert f'{"hello"!r}' == "'hello'", 'conversion !r'

# conversion !r on int (should be same as str for int)
assert f'{42!r}' == '42', 'conversion !r on int'

# conversion !r on list
assert f'{[1, 2]!r}' == '[1, 2]', 'conversion !r on list'

# conversion !s on string (no quotes)
assert f'{"hello"!s}' == 'hello', 'conversion !s on string'

# conversion !a (ascii) - escapes non-ASCII characters
assert f'{"café"!a}' == "'caf\\xe9'", 'conversion !a'
assert f'{"hello"!a}' == "'hello'", 'conversion !a ascii only'
assert f'{"日本"!a}' == "'\\u65e5\\u672c'", 'conversion !a unicode'

# === String padding and alignment ===
# format spec: width (left-aligned by default for strings)
assert f'{"hi":10}' == 'hi        ', 'format width'

# format spec: left align
assert f'{"hi":<10}' == 'hi        ', 'format left align'

# format spec: right align
assert f'{"hi":>10}' == '        hi', 'format right align'

# format spec: center align
assert f'{"hi":^10}' == '    hi    ', 'format center align'

# center align with odd padding
assert f'{"zip":^6}' == ' zip  ', 'format center align odd'

# format spec: fill character
assert f'{"hi":*>10}' == '********hi', 'format fill right'
assert f'{"hi":_<10}' == 'hi________', 'format fill left'
assert f'{"hi":*^10}' == '****hi****', 'format fill center'

# string truncation with precision
assert f'{"xylophone":.5}' == 'xylop', 'string truncation'
assert f'{"xylophone":10.5}' == 'xylop     ', 'string truncation with width'

# === Integer formatting ===
# basic integer
assert f'{42}' == '42', 'basic integer'

# integer with :d type
assert f'{42:d}' == '42', 'integer :d'

# integer padding
assert f'{42:4d}' == '  42', 'integer padding'
assert f'{42:04d}' == '0042', 'integer zero padding'

# integer with sign
assert f'{42:+d}' == '+42', 'integer positive sign'
assert f'{42: d}' == ' 42', 'integer space for positive'
assert f'{-42:+d}' == '-42', 'integer negative with sign'
assert f'{-42: d}' == '-42', 'integer negative space'

# sign-aware padding
assert f'{-23:=5d}' == '-  23', 'sign-aware padding'
assert f'{1000000:,}' == '1,000,000', 'integer comma grouping'
assert f'{255:#x}' == '0xff', 'hex alternate form'
assert f'{8:#o}' == '0o10', 'octal alternate form'
assert f'{10:#b}' == '0b1010', 'binary alternate form'

# === Float formatting ===
# basic float
assert f'{3.14159}' == '3.14159', 'basic float'

# float with :f type
assert f'{3.141592653589793:f}' == '3.141593', 'float :f'

# float precision
assert f'{3.141592653589793:.2f}' == '3.14', 'float precision'
assert f'{3.141592653589793:.4f}' == '3.1416', 'float precision 4'

# float width and precision
assert f'{3.141592653589793:06.2f}' == '003.14', 'float zero pad with precision'
assert f'{3.141592653589793:10.2f}' == '      3.14', 'float width with precision'

# float with sign
assert f'{3.14:+.2f}' == '+3.14', 'float positive sign'
assert f'{-3.14:+.2f}' == '-3.14', 'float negative with sign'
assert f'{3.14:-.2f}' == '3.14', 'float explicit minus sign'
assert f'{-3.14:-.2f}' == '-3.14', 'float explicit minus sign negative'

# exponential notation
assert f'{1234.5678:e}' == '1.234568e+03', 'exponential lowercase'
assert f'{1234.5678:E}' == '1.234568E+03', 'exponential uppercase'
assert f'{1234.5678:.2e}' == '1.23e+03', 'exponential with precision'
assert f'{0.00012345:.2e}' == '1.23e-04', 'exponential small number'

# general format (g/G) - uses exponential for very large/small numbers
assert f'{1.5:g}' == '1.5', 'general format simple'
assert f'{1.500:g}' == '1.5', 'general format strips trailing zeros'
assert f'{1234567890:g}' == '1.23457e+09', 'general format large number'

# percentage
assert f'{0.25:%}' == '25.000000%', 'percentage default precision'
assert f'{0.25:.1%}' == '25.0%', 'percentage with precision'
assert f'{0.125:.0%}' == '12%', 'percentage zero precision'

# === Nested format specs ===
width = 10
assert f'{"hi":{width}}' == 'hi        ', 'nested format spec width'

# nested alignment and width
align = '^'
assert f'{"test":{align}{width}}' == '   test   ', 'nested align and width'

# nested precision
prec = 3
assert f'{"xylophone":.{prec}}' == 'xyl', 'nested precision'


# === f-string in function ===
def greet(name):
    return f'Hello, {name}!'


assert greet('World') == 'Hello, World!', 'f-string in function'


# function returning formatted value
def format_num(n, w):
    return f'{n:>{w}}'


assert format_num('x', 5) == '    x', 'f-string with params'

# === Escaping ===
# double braces to escape
assert f'{{}}' == '{}', 'escaped braces'
assert f'{{x}}' == '{x}', 'escaped braces with content'
assert f'{{{42}}}' == '{42}', 'value inside escaped braces'

# === Complex expressions ===
# TODO: method call on literal - parser doesn't support this yet
# assert f'{"hello".upper()}' == 'HELLO', 'method call on literal'

# TODO: method call on variable - str.upper() not implemented yet
# s = 'hello'
# assert f'{s.upper()}' == 'HELLO', 'method call on variable'

# subscript in f-string
lst = [10, 20, 30]
assert f'{lst[1]}' == '20', 'subscript'

# dict lookup
d = {'a': 1, 'b': 2}
assert f'{d["a"]}' == '1', 'dict lookup'

# TODO: conditional expression - parser doesn't support IfExp yet
# x = 5
# assert f'{x if x > 0 else -x}' == '5', 'conditional positive'
# x = -5
# assert f'{-x if x < 0 else x}' == '5', 'conditional negative'

# === String concatenation ===
name = 'world'
# regular string + f-string (implicit concatenation)
assert f'hello {name}' == 'hello world', 'str concat with fstring'

# === Empty interpolation expression ===
# (this should be a syntax error, but test current behavior)
# assert f'{}' would be syntax error

# === Whitespace in format spec ===
# no extra whitespace handling needed, width handles it
assert f'{"x":5}' == 'x    ', 'single char width'

# === Unicode character counting in padding ===
x = 'café'
assert f'{x:_<10}' == 'café______'
assert f'{x:_>10}' == '______café'
assert f'{x:_^10}' == '___café___'
assert f'{x:_^11}' == '___café____'
assert f'{x:é<10}' == 'cafééééééé'
assert f'{x:é>10}' == 'éééééécafé'
assert f'{x:é^10}' == 'ééécaféééé'
assert f'{x:é^11}' == 'ééécafééééé'

# === Conversion flag with type spec ===
# conversion flag produces string, so 's' format should work
assert f'{42!r:s}' == '42', 'conversion with type spec'

# === Zero-padding with negative numbers ===
# zero-padding should use sign-aware alignment
x = -42
assert f'{x:05d}' == '-0042', 'zero pad negative'

# === Debug/self-documenting expressions (=) ===
a = 42
assert f'{a=}' == 'a=42', 'basic debug expression'
assert f'{a = }' == 'a = 42', 'debug with spaces'
name = 'test'
assert f'{name=}' == "name='test'", 'debug uses repr for strings'
assert f'{name = }' == "name = 'test'", 'debug uses repr for strings'
assert f'{name=!s}' == 'name=test', 'debug with !s conversion'
assert f'{name=!r}' == "name='test'", 'debug with !r conversion'
assert f'{1+1=}' == '1+1=2', 'debug with expression'



import textwrap

# === shorten: text fits within width ===
assert textwrap.shorten('Hello World', 12, placeholder='[...]') == 'Hello World', 'shorten fits'
assert textwrap.shorten('Hello World', 11, placeholder='[...]') == 'Hello World', 'shorten exact fit'
assert textwrap.shorten('abc', 10, placeholder='[...]') == 'abc', 'shorten short text'
assert textwrap.shorten('', 10, placeholder='[...]') == '', 'shorten empty string'

# === shorten: text truncated ===
assert textwrap.shorten('Hello World', 10, placeholder='[...]') == 'Hello[...]', 'shorten truncated'
assert textwrap.shorten('Hello World', 5, placeholder='[...]') == '[...]', 'shorten placeholder only'

# === shorten: with custom placeholder ===
assert textwrap.shorten('Hello World', 10, placeholder='...') == 'Hello...', 'shorten custom placeholder'
assert textwrap.shorten('Hello World', 3, placeholder='...') == '...', 'shorten placeholder equals width'

# === shorten: whitespace collapsing ===
assert textwrap.shorten('  Hello   World  ', 11, placeholder='[...]') == 'Hello World', 'shorten collapses whitespace'

# === shorten: longer text ===
assert textwrap.shorten('The quick brown fox', 10, placeholder='[...]') == 'The[...]', 'shorten longer text'
assert textwrap.shorten('one two three', 10, placeholder='[...]') == 'one[...]', 'shorten word boundary'

# === TextWrapper: basic wrapping ===
wrapper = textwrap.TextWrapper(width=7)
assert wrapper.wrap('one two three') == ['one two', 'three'], 'TextWrapper.wrap should split on width'
assert wrapper.fill('one two three') == 'one two\nthree', 'TextWrapper.fill should join wrapped lines with newlines'

# === TextWrapper: indentation options ===
indented = textwrap.TextWrapper(width=9, initial_indent='> ', subsequent_indent='  ')
assert indented.wrap('one two three') == ['> one two', '  three'], 'TextWrapper should apply initial/subsequent indent'

# === TextWrapper: break_long_words ===
split_words = textwrap.TextWrapper(width=5, break_long_words=True)
assert split_words.wrap('abcdefgh') == ['abcde', 'fgh'], 'TextWrapper should split long words when enabled'

keep_words = textwrap.TextWrapper(width=5, break_long_words=False)
assert keep_words.wrap('abcdefgh') == ['abcdefgh'], 'TextWrapper should keep long words whole when disabled'

# === TextWrapper: exposed attributes ===
configured = textwrap.TextWrapper(
    width=12,
    initial_indent='>>',
    subsequent_indent='..',
    placeholder='***',
    max_lines=3,
    break_long_words=False,
)
assert configured.width == 12, 'TextWrapper.width should reflect constructor arg'
assert configured.initial_indent == '>>', 'TextWrapper.initial_indent should reflect constructor arg'
assert configured.subsequent_indent == '..', 'TextWrapper.subsequent_indent should reflect constructor arg'
assert configured.placeholder == '***', 'TextWrapper.placeholder should reflect constructor arg'
assert configured.max_lines == 3, 'TextWrapper.max_lines should reflect constructor arg'



from typing import Any

import pytest
from inline_snapshot import snapshot

import ouros


def test_start_no_external_functions_returns_complete():
    m = ouros.Sandbox('1 + 2')
    result = m.start()
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(3)


def test_start_with_external_function_returns_progress():
    m = ouros.Sandbox('func()', external_functions=['func'])
    result = m.start()
    assert isinstance(result, ouros.Snapshot)
    assert result.script_name == snapshot('main.py')
    assert result.function_name == snapshot('func')
    assert result.args == snapshot(())
    assert result.kwargs == snapshot({})


def test_start_custom_script_name():
    m = ouros.Sandbox('func()', script_name='custom.py', external_functions=['func'])
    result = m.start()
    assert isinstance(result, ouros.Snapshot)
    assert result.script_name == snapshot('custom.py')


def test_start_progress_resume_returns_complete():
    m = ouros.Sandbox('func()', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('func')
    assert progress.args == snapshot(())
    assert progress.kwargs == snapshot({})

    result = progress.resume(return_value=42)
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(42)


def test_start_progress_with_args():
    m = ouros.Sandbox('func(1, 2, 3)', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('func')
    assert progress.args == snapshot((1, 2, 3))
    assert progress.kwargs == snapshot({})


def test_start_progress_with_kwargs():
    m = ouros.Sandbox('func(a=1, b="two")', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('func')
    assert progress.args == snapshot(())
    assert progress.kwargs == snapshot({'a': 1, 'b': 'two'})


def test_start_progress_with_mixed_args_kwargs():
    m = ouros.Sandbox('func(1, 2, x="hello", y=True)', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('func')
    assert progress.args == snapshot((1, 2))
    assert progress.kwargs == snapshot({'x': 'hello', 'y': True})


def test_start_multiple_external_calls():
    m = ouros.Sandbox('a() + b()', external_functions=['a', 'b'])

    # First call
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('a')

    # Resume with first return value
    progress = progress.resume(return_value=10)
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('b')

    # Resume with second return value
    result = progress.resume(return_value=5)
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(15)


def test_start_chain_of_external_calls():
    m = ouros.Sandbox('c() + c() + c()', external_functions=['c'])

    call_count = 0
    progress: ouros.Snapshot | ouros.FutureSnapshot | ouros.Complete = m.start()

    while isinstance(progress, ouros.Snapshot | ouros.FutureSnapshot):
        assert isinstance(progress, ouros.Snapshot), 'Expected Snapshot'
        assert progress.function_name == snapshot('c')
        call_count += 1
        progress = progress.resume(return_value=call_count)

    assert isinstance(progress, ouros.Complete)
    assert progress.output == snapshot(6)  # 1 + 2 + 3
    assert call_count == snapshot(3)


def test_start_with_inputs():
    m = ouros.Sandbox('process(x)', inputs=['x'], external_functions=['process'])
    progress = m.start(inputs={'x': 100})
    assert isinstance(progress, ouros.Snapshot)
    assert progress.function_name == snapshot('process')
    assert progress.args == snapshot((100,))


def test_start_with_limits():
    m = ouros.Sandbox('1 + 2')
    limits = ouros.ResourceLimits(max_allocations=1000)
    result = m.start(limits=limits)
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(3)


def test_start_with_print_callback():
    output: list[tuple[str, str]] = []

    def callback(stream: str, text: str) -> None:
        output.append((stream, text))

    m = ouros.Sandbox('print("hello")')
    result = m.start(print_callback=callback)
    assert isinstance(result, ouros.Complete)
    assert output == snapshot([('stdout', "hello\n")])


def test_start_resume_cannot_be_called_twice():
    m = ouros.Sandbox('func()', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    # First resume succeeds
    progress.resume(return_value=1)

    # Second resume should fail
    with pytest.raises(RuntimeError) as exc_info:
        progress.resume(return_value=2)
    assert exc_info.value.args[0] == snapshot('Progress already resumed')


def test_start_complex_return_value():
    m = ouros.Sandbox('func()', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    result = progress.resume(return_value={'a': [1, 2, 3], 'b': {'nested': True}})
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot({'a': [1, 2, 3], 'b': {'nested': True}})


def test_start_resume_with_none():
    m = ouros.Sandbox('func()', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    result = progress.resume(return_value=None)
    assert isinstance(result, ouros.Complete)
    assert result.output is None


def test_progress_repr():
    m = ouros.Sandbox('func(1, x=2)', external_functions=['func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    assert repr(progress) == snapshot(
        "Snapshot(script_name='main.py', function_name='func', args=(1,), kwargs={'x': 2})"
    )


def test_complete_repr():
    m = ouros.Sandbox('42')
    result = m.start()
    assert isinstance(result, ouros.Complete)
    assert repr(result) == snapshot('Complete(output=42)')


def test_start_can_reuse_ouros_instance():
    m = ouros.Sandbox('func(x)', inputs=['x'], external_functions=['func'])

    # First run
    progress1 = m.start(inputs={'x': 1})
    assert isinstance(progress1, ouros.Snapshot)
    assert progress1.args == snapshot((1,))
    result1 = progress1.resume(return_value=10)
    assert isinstance(result1, ouros.Complete)
    assert result1.output == snapshot(10)

    # Second run with different input
    progress2 = m.start(inputs={'x': 2})
    assert isinstance(progress2, ouros.Snapshot)
    assert progress2.args == snapshot((2,))
    result2 = progress2.resume(return_value=20)
    assert isinstance(result2, ouros.Complete)
    assert result2.output == snapshot(20)


@pytest.mark.parametrize(
    'code,expected',
    [
        ('1', 1),
        ('"hello"', 'hello'),
        ('[1, 2, 3]', [1, 2, 3]),
        ('{"a": 1}', {'a': 1}),
        ('None', None),
        ('True', True),
    ],
)
def test_start_returns_complete_for_various_types(code: str, expected: Any):
    m = ouros.Sandbox(code)
    result = m.start()
    assert isinstance(result, ouros.Complete)
    assert result.output == expected


def test_start_progress_resume_with_exception_caught():
    """Test that resuming with an exception is caught by try/except."""
    code = """
try:
    result = external_func()
except ValueError:
    caught = True
caught
"""
    m = ouros.Sandbox(code, external_functions=['external_func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    # Resume with an exception using keyword argument
    result = progress.resume(exception=ValueError('test error'))
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(True)


def test_start_progress_resume_exception_propagates_uncaught():
    """Test that uncaught exceptions from resume() propagate to caller."""
    code = 'external_func()'
    m = ouros.Sandbox(code, external_functions=['external_func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    # Resume with an exception that won't be caught - wrapped in SandboxRuntimeError
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        progress.resume(exception=ValueError('uncaught error'))
    inner = exc_info.value.exception()
    assert isinstance(inner, ValueError)
    assert inner.args[0] == snapshot('uncaught error')


def test_resume_none():
    code = 'external_func()'
    m = ouros.Sandbox(code, external_functions=['external_func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)
    result = progress.resume(return_value=None)
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot(None)


def test_invalid_resume_args():
    """Test that resume() with no args returns None."""
    code = 'external_func()'
    m = ouros.Sandbox(code, external_functions=['external_func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    # no args provided
    with pytest.raises(TypeError) as exc_info:
        progress.resume()  # pyright: ignore[reportCallIssue]
    assert exc_info.value.args[0] == snapshot('resume() accepts either return_value or exception, not both')

    # Both arguments provided
    with pytest.raises(TypeError) as exc_info:
        progress.resume(return_value=42, exception=ValueError('error'))  # pyright: ignore[reportCallIssue]
    assert exc_info.value.args[0] == snapshot('resume() accepts either return_value or exception, not both')

    # invalid kwarg provided
    with pytest.raises(TypeError) as exc_info:
        progress.resume(invalid_kwarg=42)  # pyright: ignore[reportCallIssue]
    assert exc_info.value.args[0] == snapshot('resume() accepts either return_value or exception, not both')


def test_start_progress_resume_exception_in_nested_try():
    """Test exception handling in nested try/except blocks."""
    code = """
outer_caught = False
finally_ran = False
try:
    try:
        external_func()
    except TypeError:
        pass  # Won't catch ValueError
    finally:
        finally_ran = True
except ValueError:
    outer_caught = True
(outer_caught, finally_ran)
"""
    m = ouros.Sandbox(code, external_functions=['external_func'])
    progress = m.start()
    assert isinstance(progress, ouros.Snapshot)

    result = progress.resume(exception=ValueError('propagates to outer'))
    assert isinstance(result, ouros.Complete)
    assert result.output == snapshot((True, True))



# call-external
import dataclasses

# === is_dataclass ===
# Returns True for dataclass instances, False for everything else
point = make_point()
assert dataclasses.is_dataclass(point) == True, 'is_dataclass on frozen dataclass is True'

mut_point = make_mutable_point()
assert dataclasses.is_dataclass(mut_point) == True, 'is_dataclass on mutable dataclass is True'

alice = make_user('Alice')
assert dataclasses.is_dataclass(alice) == True, 'is_dataclass on user dataclass is True'

empty = make_empty()
assert dataclasses.is_dataclass(empty) == True, 'is_dataclass on empty dataclass is True'

# Non-dataclass values
assert dataclasses.is_dataclass(42) == False, 'is_dataclass on int is False'
assert dataclasses.is_dataclass('hello') == False, 'is_dataclass on str is False'
assert dataclasses.is_dataclass(None) == False, 'is_dataclass on None is False'
assert dataclasses.is_dataclass([1, 2]) == False, 'is_dataclass on list is False'
assert dataclasses.is_dataclass({}) == False, 'is_dataclass on dict is False'
assert dataclasses.is_dataclass(True) == False, 'is_dataclass on bool is False'
assert dataclasses.is_dataclass((1,)) == False, 'is_dataclass on tuple is False'

# === fields ===
# Returns field descriptors; both CPython and Ouros agree on the count
f_point = dataclasses.fields(point)
assert len(f_point) == 2, 'frozen point has 2 fields'

f_user = dataclasses.fields(alice)
assert len(f_user) == 2, 'user has 2 fields'

f_empty = dataclasses.fields(empty)
assert len(f_empty) == 0, 'empty has 0 fields'

f_mut = dataclasses.fields(mut_point)
assert len(f_mut) == 2, 'mutable point has 2 fields'

# === asdict ===
# Converts dataclass instance to dict
d_point = dataclasses.asdict(point)
assert d_point == {'x': 1, 'y': 2}, 'asdict frozen point'
assert type(d_point) is dict, 'asdict returns dict type'

d_user = dataclasses.asdict(alice)
assert d_user == {'name': 'Alice', 'active': True}, 'asdict user with string field'

d_empty = dataclasses.asdict(empty)
assert d_empty == {}, 'asdict empty dataclass returns empty dict'

d_mut = dataclasses.asdict(mut_point)
assert d_mut == {'x': 1, 'y': 2}, 'asdict mutable point'

# === astuple ===
# Converts dataclass instance to tuple in field order
t_point = dataclasses.astuple(point)
assert t_point == (1, 2), 'astuple frozen point'
assert type(t_point) is tuple, 'astuple returns tuple type'

t_user = dataclasses.astuple(alice)
assert t_user == ('Alice', True), 'astuple user'

t_empty = dataclasses.astuple(empty)
assert t_empty == (), 'astuple empty dataclass returns empty tuple'

t_mut = dataclasses.astuple(mut_point)
assert t_mut == (1, 2), 'astuple mutable point'

# === replace on mutable dataclass ===
# Creates a shallow copy with specified field overrides
mut2 = make_mutable_point()
replaced = dataclasses.replace(mut2, x=99)
assert replaced.x == 99, 'replace overrides x'
assert replaced.y == 2, 'replace preserves y'
assert mut2.x == 1, 'original unchanged after replace'

# replace with multiple overrides
replaced2 = dataclasses.replace(mut2, x=100, y=200)
assert replaced2.x == 100, 'replace both x'
assert replaced2.y == 200, 'replace both y'
assert mut2.x == 1, 'original x still unchanged'

# replace with no changes (copy)
copied = dataclasses.replace(mut2)
assert copied.x == 1, 'copy x matches original'
assert copied.y == 2, 'copy y matches original'

# === replace on frozen dataclass ===
point2 = make_point()
replaced_frozen = dataclasses.replace(point2, x=50)
assert replaced_frozen.x == 50, 'replace on frozen overrides x'
assert replaced_frozen.y == 2, 'replace on frozen preserves y'
assert point2.x == 1, 'original frozen point unchanged'

# === replace preserves is_dataclass ===
assert dataclasses.is_dataclass(replaced) == True, 'replaced mutable is still dataclass'
assert dataclasses.is_dataclass(replaced_frozen) == True, 'replaced frozen is still dataclass'

# === asdict and astuple on replaced instance ===
d_replaced = dataclasses.asdict(replaced)
assert d_replaced == {'x': 99, 'y': 2}, 'asdict on replaced instance'

t_replaced = dataclasses.astuple(replaced)
assert t_replaced == (99, 2), 'astuple on replaced instance'

# === replace on user dataclass ===
bob = make_user('Bob')
replaced_user = dataclasses.replace(bob, name='Charlie')
assert replaced_user.name == 'Charlie', 'replace user name'
assert replaced_user.active == True, 'replace preserves active'
assert bob.name == 'Bob', 'original user unchanged'

# === make_dataclass ===
dynamic_point_factory = dataclasses.make_dataclass('DynamicPoint', ['x', 'y'])
dynamic_user_factory = dataclasses.make_dataclass('DynamicUser', [('name', str), ('age', int)])
dynamic_status_factory = dataclasses.make_dataclass('DynamicStatus', [['active', bool], 'score'])

if hasattr(dynamic_point_factory, '__mro__'):
    # CPython behavior: make_dataclass returns a class.
    dynamic_point = dynamic_point_factory(x=1, y=2)
    point_field_names = [f.name for f in dataclasses.fields(dynamic_point_factory)]
    assert point_field_names == ['x', 'y'], 'make_dataclass should preserve field order for string specs'
    assert dataclasses.asdict(dynamic_point) == {'x': 1, 'y': 2}, 'make_dataclass class instances should support asdict'

    dynamic_user = dynamic_user_factory(name='Ada', age=36)
    user_field_names = [f.name for f in dataclasses.fields(dynamic_user_factory)]
    assert user_field_names == ['name', 'age'], 'make_dataclass should support tuple field specs'

    dynamic_status = dynamic_status_factory(active=True, score=7)
    status_field_names = [f.name for f in dataclasses.fields(dynamic_status_factory)]
    assert status_field_names == ['active', 'score'], 'make_dataclass should support list field specs'
    assert dataclasses.astuple(dynamic_status) == (True, 7), 'list field specs should preserve tuple field order'

    updated_dynamic_user = dataclasses.replace(dynamic_user, age=37)
    assert dataclasses.asdict(updated_dynamic_user) == {'name': 'Ada', 'age': 37}, (
        'replace should work on make_dataclass instances'
    )
    assert dataclasses.asdict(dynamic_user) == {'name': 'Ada', 'age': 36}, (
        'replace should not mutate make_dataclass original'
    )
else:
    # Ouros behavior: make_dataclass returns a dataclass-like instance.
    dynamic_point = dynamic_point_factory
    dynamic_user = dynamic_user_factory
    dynamic_status = dynamic_status_factory

    assert dataclasses.is_dataclass(dynamic_point) == True, 'make_dataclass should return dataclass-like values'
    assert [f.name for f in dataclasses.fields(dynamic_point)] == ['x', 'y'], (
        'make_dataclass should preserve field order for string specs'
    )
    assert dataclasses.asdict(dynamic_point) == {'x': None, 'y': None}, (
        'make_dataclass should initialize fields to None'
    )
    assert [f.name for f in dataclasses.fields(dynamic_user)] == ['name', 'age'], 'make_dataclass should support tuple field specs'
    assert dataclasses.asdict(dynamic_user) == {'name': None, 'age': None}, (
        'tuple field specs should initialize to None'
    )
    assert [f.name for f in dataclasses.fields(dynamic_status)] == ['active', 'score'], 'make_dataclass should support list field specs'
    assert dataclasses.astuple(dynamic_status) == (None, None), 'list field specs should preserve tuple field order'

    updated_dynamic_user = dataclasses.replace(dynamic_user, name='Ada', age=36)
    assert dataclasses.asdict(updated_dynamic_user) == {'name': 'Ada', 'age': 36}, (
        'replace should work on make_dataclass output'
    )
    assert dataclasses.asdict(dynamic_user) == {'name': None, 'age': None}, (
        'replace should not mutate make_dataclass original'
    )

# === make_dataclass and replace error cases ===
duplicate_field_error = False
try:
    dataclasses.make_dataclass('DupField', ['x', ('x', int)])
except TypeError:
    duplicate_field_error = True
assert duplicate_field_error, 'make_dataclass should reject duplicate field names'

replace_positional_error = False
try:
    dataclasses.replace(dynamic_user, dynamic_point)
except TypeError:
    replace_positional_error = True
assert replace_positional_error, 'replace should reject extra positional arguments'

# === decorator parity: ClassVar / InitVar / tuple annotations ===
from typing import ClassVar


@dataclasses.dataclass
class ClassVarInitVarCase:
    coords: tuple[int, int]
    token: dataclasses.InitVar[int]
    category: ClassVar[str] = 'meta'
    seen: int = 0

    def __post_init__(self, token):
        self.seen = token


case = ClassVarInitVarCase((1, 2), 7)
case_fields = dataclasses.fields(ClassVarInitVarCase)
case_field_names = [f.name for f in case_fields]
assert case_field_names == ['coords', 'seen'], 'fields() should exclude InitVar and ClassVar'
assert dataclasses.asdict(case) == {'coords': (1, 2), 'seen': 7}, 'asdict should exclude InitVar/ClassVar fields'
assert dataclasses.astuple(case) == ((1, 2), 7), 'astuple should preserve tuple-typed field and exclude InitVar/ClassVar'
assert 'token' not in case.__dict__, 'InitVar should not become an instance attribute'
assert ClassVarInitVarCase.category == 'meta', 'ClassVar value should remain a class attribute'

# === mutable default rejection ===
mutable_list_default_error = False
try:
    @dataclasses.dataclass
    class MutableListDefault:
        xs: list = []
except ValueError:
    mutable_list_default_error = True
assert mutable_list_default_error, 'mutable list defaults should require default_factory'

mutable_dict_default_error = False
try:
    @dataclasses.dataclass
    class MutableDictDefault:
        xs: dict = {}
except ValueError:
    mutable_dict_default_error = True
assert mutable_dict_default_error, 'mutable dict defaults should require default_factory'

# === dataclass comparison dunders should return NotImplemented for mismatched types ===
@dataclasses.dataclass(order=True)
class OrderedLeft:
    value: int


@dataclasses.dataclass(order=True)
class OrderedRight:
    value: int


ordered_left = OrderedLeft(1)
ordered_right = OrderedRight(1)
assert OrderedLeft.__eq__(ordered_left, ordered_right) is NotImplemented, '__eq__ should return NotImplemented on type mismatch'
assert OrderedLeft.__lt__(ordered_left, ordered_right) is NotImplemented, '__lt__ should return NotImplemented on type mismatch'
assert (ordered_left == ordered_right) == False, 'equality should still evaluate to False for mismatched dataclass types'

# === replace() parity for InitVar and init=False fields ===
@dataclasses.dataclass
class ReplaceParityCase:
    x: int
    token: dataclasses.InitVar[int]
    y: int = dataclasses.field(init=False, default=0)

    def __post_init__(self, token):
        self.y = token


replace_base = ReplaceParityCase(1, 5)
replace_requires_initvar = False
try:
    dataclasses.replace(replace_base, x=2)
except TypeError:
    replace_requires_initvar = True
assert replace_requires_initvar, 'replace() should require InitVar values without defaults'

replace_init_false_error = False
try:
    dataclasses.replace(replace_base, x=2, token=9, y=3)
except TypeError:
    replace_init_false_error = True
assert replace_init_false_error, 'replace() should reject fields declared with init=False'

replace_ok = dataclasses.replace(replace_base, x=2, token=9)
assert dataclasses.asdict(replace_ok) == {'x': 2, 'y': 9}, 'replace() should rebuild non-init fields via __post_init__'



# call-external
# BUG: This test hangs (infinite loop) - external calls in boolean expressions
# with side effects cause incorrect behavior.

call_count = 0


def side_effect(val):
    global call_count
    call_count += 1
    return val


# This specific pattern causes a hang in Ouros
result = return_value(True) and return_value(side_effect(42))
assert result == 42, 'and: second runs when first truthy'
assert call_count == 1, 'and: side effect runs when first is truthy'



# === min() ===
# Basic min operations
assert min([1, 2, 3]) == 1, 'min of list'
assert min([3, 1, 2]) == 1, 'min of unsorted list'
assert min([5]) == 5, 'min of single element'
assert min(1, 2, 3) == 1, 'min of multiple args'
assert min(3, 1, 2) == 1, 'min of unsorted args'
assert min(-5, -10, -1) == -10, 'min of negatives'

# min with strings
assert min(['b', 'a', 'c']) == 'a', 'min of string list'
assert min('b', 'a', 'c') == 'a', 'min of string args'

# min with floats
assert min([1.5, 0.5, 2.5]) == 0.5, 'min of floats'
assert min(1.5, 0.5) == 0.5, 'min float args'

# min with key
assert min(['bbb', 'a', 'cc'], key=len) == 'a', 'min iterable with key'
assert min('bbb', 'a', 'cc', key=len) == 'a', 'min variadic args with key'
assert min(['bbb', 'a', 'cc'], key=lambda s: len(s)) == 'a', 'min iterable with lambda key'
assert min('bbb', 'a', 'cc', key=lambda s: len(s)) == 'a', 'min variadic args with lambda key'

# === max() ===
# Basic max operations
assert max([1, 2, 3]) == 3, 'max of list'
assert max([3, 1, 2]) == 3, 'max of unsorted list'
assert max([5]) == 5, 'max of single element'
assert max(1, 2, 3) == 3, 'max of multiple args'
assert max(3, 1, 2) == 3, 'max of unsorted args'
assert max(-5, -10, -1) == -1, 'max of negatives'

# max with strings
assert max(['b', 'a', 'c']) == 'c', 'max of string list'
assert max('b', 'a', 'c') == 'c', 'max of string args'

# max with floats
assert max([1.5, 0.5, 2.5]) == 2.5, 'max of floats'
assert max(1.5, 2.5) == 2.5, 'max float args'

# max with key
assert max(['bbb', 'a', 'cc'], key=len) == 'bbb', 'max iterable with key'
assert max('bbb', 'a', 'cc', key=len) == 'bbb', 'max variadic args with key'
assert max(['bbb', 'a', 'cc'], key=lambda s: len(s)) == 'bbb', 'max iterable with lambda key'
assert max('bbb', 'a', 'cc', key=lambda s: len(s)) == 'bbb', 'max variadic args with lambda key'

# === sorted() ===
# Basic sorted operations
assert sorted([3, 1, 2]) == [1, 2, 3], 'sorted int list'
assert sorted([1, 2, 3]) == [1, 2, 3], 'sorted already sorted'
assert sorted([3, 2, 1]) == [1, 2, 3], 'sorted reverse order'
assert sorted([]) == [], 'sorted empty list'
assert sorted([5]) == [5], 'sorted single element'

# sorted with strings
assert sorted(['c', 'a', 'b']) == ['a', 'b', 'c'], 'sorted strings'

# sorted with negative numbers
assert sorted([-3, 1, -2, 2]) == [-3, -2, 1, 2], 'sorted with negatives'

# sorted with tuple
assert sorted((3, 1, 2)) == [1, 2, 3], 'sorted tuple returns list'

# sorted preserves duplicates
assert sorted([3, 1, 2, 1, 3]) == [1, 1, 2, 3, 3], 'sorted with duplicates'

# sorted with range
assert sorted(range(5, 0, -1)) == [1, 2, 3, 4, 5], 'sorted range'

# === reversed() ===
# Basic reversed operations
assert list(reversed([1, 2, 3])) == [3, 2, 1], 'reversed list'
assert list(reversed([1])) == [1], 'reversed single element'
assert list(reversed([])) == [], 'reversed empty list'

# reversed tuple
assert list(reversed((1, 2, 3))) == [3, 2, 1], 'reversed tuple'

# reversed string
assert list(reversed('abc')) == ['c', 'b', 'a'], 'reversed string'

# reversed range
assert list(reversed(range(1, 4))) == [3, 2, 1], 'reversed range'

# === enumerate() ===
# Basic enumerate operations
assert list(enumerate(['a', 'b', 'c'])) == [(0, 'a'), (1, 'b'), (2, 'c')], 'enumerate list'
assert list(enumerate([])) == [], 'enumerate empty list'
assert list(enumerate(['x'])) == [(0, 'x')], 'enumerate single element'

# enumerate with start
assert list(enumerate(['a', 'b'], 1)) == [(1, 'a'), (2, 'b')], 'enumerate with start'
assert list(enumerate(['a', 'b'], 10)) == [(10, 'a'), (11, 'b')], 'enumerate with start 10'

# enumerate string
assert list(enumerate('ab')) == [(0, 'a'), (1, 'b')], 'enumerate string'

# enumerate range
assert list(enumerate(range(3))) == [(0, 0), (1, 1), (2, 2)], 'enumerate range'

# === zip() ===
# Basic zip operations
assert list(zip([1, 2], ['a', 'b'])) == [(1, 'a'), (2, 'b')], 'zip two lists'
assert list(zip([1], ['a'])) == [(1, 'a')], 'zip single elements'
assert list(zip([], [])) == [], 'zip empty lists'

# zip truncates to shortest
assert list(zip([1, 2, 3], ['a', 'b'])) == [(1, 'a'), (2, 'b')], 'zip truncates to shortest'
assert list(zip([1], ['a', 'b', 'c'])) == [(1, 'a')], 'zip truncates first shorter'

# zip three iterables
assert list(zip([1, 2], ['a', 'b'], [True, False])) == [(1, 'a', True), (2, 'b', False)], 'zip three lists'

# zip with different types
assert list(zip(range(3), 'abc')) == [(0, 'a'), (1, 'b'), (2, 'c')], 'zip range and string'

# zip single iterable
assert list(zip([1, 2, 3])) == [(1,), (2,), (3,)], 'zip single iterable'

# zip with empty
assert list(zip([1, 2], [])) == [], 'zip with empty second'
assert list(zip([], [1, 2])) == [], 'zip with empty first'



def foo():
    continue


foo()
"""
TRACEBACK:
Traceback (most recent call last):
  File "loop__continue_in_function_error.py", line 2
    continue
    ~~~~~~~~
SyntaxError: 'continue' not properly in loop
"""



# Tests for error cases in BigInt-related builtins and operations
# All error messages must match CPython exactly
# Uses 'in str(e)' checks since Ouros's str(e) includes the type name

# === Setup constants ===
MAX_I64 = 9223372036854775807
BIGINT = MAX_I64 + 1  # Force BigInt creation


# === hex() errors ===
try:
    hex('str')
    assert False, 'hex(str) should raise TypeError'
except TypeError as e:
    assert "'str' object cannot be interpreted as an integer" in str(e), f'hex str error: {e}'

try:
    hex(1.5)
    assert False, 'hex(float) should raise TypeError'
except TypeError as e:
    assert "'float' object cannot be interpreted as an integer" in str(e), f'hex float error: {e}'

try:
    hex([])
    assert False, 'hex(list) should raise TypeError'
except TypeError as e:
    assert "'list' object cannot be interpreted as an integer" in str(e), f'hex list error: {e}'


# === bin() errors ===
try:
    bin('str')
    assert False, 'bin(str) should raise TypeError'
except TypeError as e:
    assert "'str' object cannot be interpreted as an integer" in str(e), f'bin str error: {e}'

try:
    bin(1.5)
    assert False, 'bin(float) should raise TypeError'
except TypeError as e:
    assert "'float' object cannot be interpreted as an integer" in str(e), f'bin float error: {e}'

try:
    bin({})
    assert False, 'bin(dict) should raise TypeError'
except TypeError as e:
    assert "'dict' object cannot be interpreted as an integer" in str(e), f'bin dict error: {e}'


# === oct() errors ===
try:
    oct('str')
    assert False, 'oct(str) should raise TypeError'
except TypeError as e:
    assert "'str' object cannot be interpreted as an integer" in str(e), f'oct str error: {e}'

try:
    oct(1.5)
    assert False, 'oct(float) should raise TypeError'
except TypeError as e:
    assert "'float' object cannot be interpreted as an integer" in str(e), f'oct float error: {e}'

try:
    oct((1, 2))
    assert False, 'oct(tuple) should raise TypeError'
except TypeError as e:
    assert "'tuple' object cannot be interpreted as an integer" in str(e), f'oct tuple error: {e}'


# === divmod() division by zero ===
try:
    divmod(10, 0)
    assert False, 'divmod(int, 0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod int/0 error: {e}'

try:
    divmod(BIGINT, 0)
    assert False, 'divmod(bigint, 0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod bigint/0 error: {e}'

try:
    divmod(10, BIGINT - BIGINT)  # BigInt zero
    assert False, 'divmod(int, bigint_zero) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod int/bigint_zero error: {e}'

try:
    divmod(BIGINT, BIGINT - BIGINT)  # BigInt / BigInt zero
    assert False, 'divmod(bigint, bigint_zero) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod bigint/bigint_zero error: {e}'

try:
    divmod(10.0, 0.0)
    assert False, 'divmod(float, 0.0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod float/0.0 error: {e}'

try:
    divmod(10, 0.0)
    assert False, 'divmod(int, 0.0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod int/0.0 error: {e}'

try:
    divmod(10.0, 0)
    assert False, 'divmod(float, 0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'divmod float/0 error: {e}'


# === divmod() type errors ===
try:
    divmod('a', 5)
    assert False, 'divmod(str, int) should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for divmod(): 'str' and 'int'" in str(e), f'divmod str/int error: {e}'

try:
    divmod(5, 'a')
    assert False, 'divmod(int, str) should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for divmod(): 'int' and 'str'" in str(e), f'divmod int/str error: {e}'

try:
    divmod([], 5)
    assert False, 'divmod(list, int) should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for divmod(): 'list' and 'int'" in str(e), f'divmod list/int error: {e}'

try:
    divmod(BIGINT, 'a')
    assert False, 'divmod(bigint, str) should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for divmod(): 'int' and 'str'" in str(e), f'divmod bigint/str error: {e}'


# === pow() zero to negative power ===
try:
    pow(0.0, -1)
    assert False, 'pow(0.0, -1) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'zero to a negative power' in str(e), f'pow 0.0/-1 error: {e}'

try:
    pow(0, -1.0)
    assert False, 'pow(0, -1.0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'zero to a negative power' in str(e), f'pow 0/-1.0 error: {e}'

try:
    pow(0.0, -2.0)
    assert False, 'pow(0.0, -2.0) should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'zero to a negative power' in str(e), f'pow 0.0/-2.0 error: {e}'


# === pow() with modulo errors ===
try:
    pow(2, 10, 0)
    assert False, 'pow(2, 10, 0) should raise ValueError'
except ValueError as e:
    assert 'pow() 3rd argument cannot be 0' in str(e), f'pow mod=0 error: {e}'

# Note: pow(2, -1, 5) computes modular inverse in Python 3.8+, not an error
# But pow(2, -1, 4) raises an error because 2 is not invertible mod 4
try:
    pow(2, -1, 4)  # gcd(2, 4) != 1, no inverse exists
    assert False, 'pow(2, -1, 4) should raise ValueError'
except ValueError as e:
    # CPython: "base is not invertible for the given modulus"
    # Ouros: "pow() 2nd argument cannot be negative when 3rd argument specified"
    # Accept either message since Ouros doesn't support modular inverse yet
    assert 'not invertible' in str(e) or 'cannot be negative' in str(e), f'pow non-invertible error: {e}'

try:
    pow(2.0, 2, 5)
    assert False, 'pow(float, int, int) should raise TypeError'
except TypeError as e:
    assert 'pow() 3rd argument not allowed unless all arguments are integers' in str(e), f'pow float mod error: {e}'

try:
    pow(2, 2.0, 5)
    assert False, 'pow(int, float, int) should raise TypeError'
except TypeError as e:
    assert 'pow() 3rd argument not allowed unless all arguments are integers' in str(e), f'pow float exp mod error: {e}'

try:
    pow(2, 2, 5.0)
    assert False, 'pow(int, int, float) should raise TypeError'
except TypeError as e:
    assert 'pow() 3rd argument not allowed unless all arguments are integers' in str(e), f'pow float mod2 error: {e}'


# === abs() type errors ===
try:
    abs('str')
    assert False, 'abs(str) should raise TypeError'
except TypeError as e:
    assert "bad operand type for abs(): 'str'" in str(e), f'abs str error: {e}'

try:
    abs([])
    assert False, 'abs(list) should raise TypeError'
except TypeError as e:
    assert "bad operand type for abs(): 'list'" in str(e), f'abs list error: {e}'

try:
    abs({})
    assert False, 'abs(dict) should raise TypeError'
except TypeError as e:
    assert "bad operand type for abs(): 'dict'" in str(e), f'abs dict error: {e}'


# === pow() type errors (** operator) ===
try:
    5 ** 'x'
    assert False, '5 ** str should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for ** or pow(): 'int' and 'str'" in str(e), f'int ** str error: {e}'

try:
    'x' ** 5
    assert False, 'str ** int should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for ** or pow(): 'str' and 'int'" in str(e), f'str ** int error: {e}'

try:
    BIGINT ** 'x'
    assert False, 'bigint ** str should raise TypeError'
except TypeError as e:
    assert "unsupported operand type(s) for ** or pow(): 'int' and 'str'" in str(e), f'bigint ** str error: {e}'


# === Division by zero with BigInt ===
try:
    BIGINT // 0
    assert False, 'bigint // 0 should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'bigint floordiv error: {e}'

try:
    BIGINT % 0
    assert False, 'bigint % 0 should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'bigint mod error: {e}'

try:
    10 // (BIGINT - BIGINT)  # int // BigInt zero
    assert False, 'int // bigint_zero should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'int floordiv bigint_zero error: {e}'

try:
    10 % (BIGINT - BIGINT)  # int % BigInt zero
    assert False, 'int % bigint_zero should raise ZeroDivisionError'
except ZeroDivisionError as e:
    assert 'division by zero' in str(e), f'int mod bigint_zero error: {e}'



# xfail=ouros
# Ouros's list += only supports other lists, not arbitrary iterables.
# CPython's list += calls extend() which requires an iterable.
x = [1]
x += 2
# Raise=TypeError("'int' object is not iterable")



# === type() 1-arg returns type ===
assert type(42) == int, 'type(42) should be int'
assert type('hi') == str, 'type(hi) should be str'

# === __init_subclass__ called ===
class Base:
    subclasses = []

    def __init_subclass__(cls, **kwargs):
        Base.subclasses.append(cls)


class Child(Base):
    pass


assert len(Base.subclasses) == 1, '__init_subclass__ should be called'
assert Base.subclasses[0] is Child, 'Child should be in subclasses'



import enum
import sre_compile
import sre_constants
import sys
from _typeshed import MaybeNone, ReadableBuffer
from collections.abc import Callable, Iterator, Mapping
from types import GenericAlias
from typing import Any, AnyStr, Final, Generic, Literal, TypeVar, final, overload
from typing_extensions import TypeAlias, deprecated

__all__ = [
    "match",
    "fullmatch",
    "search",
    "sub",
    "subn",
    "split",
    "findall",
    "finditer",
    "compile",
    "purge",
    "escape",
    "error",
    "A",
    "I",
    "L",
    "M",
    "S",
    "X",
    "U",
    "ASCII",
    "IGNORECASE",
    "LOCALE",
    "MULTILINE",
    "DOTALL",
    "VERBOSE",
    "UNICODE",
    "Match",
    "Pattern",
]
if sys.version_info < (3, 13):
    __all__ += ["template"]

if sys.version_info >= (3, 11):
    __all__ += ["NOFLAG", "RegexFlag"]

if sys.version_info >= (3, 13):
    __all__ += ["PatternError"]

    PatternError = sre_constants.error

_T = TypeVar("_T")

# The implementation defines this in re._constants (version_info >= 3, 11) or
# sre_constants. Typeshed has it here because its __module__ attribute is set to "re".
class error(Exception):
    msg: str
    pattern: str | bytes | None
    pos: int | None
    lineno: int
    colno: int
    def __init__(self, msg: str, pattern: str | bytes | None = None, pos: int | None = None) -> None: ...

@final
class Match(Generic[AnyStr]):
    @property
    def pos(self) -> int:
        """The index into the string at which the RE engine started looking for a match."""
        ...
    @property
    def endpos(self) -> int:
        """The index into the string beyond which the RE engine will not go."""
        ...
    @property
    def lastindex(self) -> int | None:
        """The integer index of the last matched capturing group."""
        ...
    @property
    def lastgroup(self) -> str | None:
        """The name of the last matched capturing group."""
        ...
    @property
    def string(self) -> AnyStr:
        """The string passed to match() or search()."""
        ...

    # The regular expression object whose match() or search() method produced
    # this match instance.
    @property
    def re(self) -> Pattern[AnyStr]:
        """The regular expression object."""
        ...
    @overload
    def expand(self: Match[str], template: str) -> str:
        """Return the string obtained by doing backslash substitution on the string template, as done by the sub() method."""
        ...
    @overload
    def expand(self: Match[bytes], template: ReadableBuffer) -> bytes:
        """Return the string obtained by doing backslash substitution on the string template, as done by the sub() method."""
        ...
    @overload
    def expand(self, template: AnyStr) -> AnyStr:
        """Return the string obtained by doing backslash substitution on the string template, as done by the sub() method."""
        ...
    # group() returns "AnyStr" or "AnyStr | None", depending on the pattern.
    @overload
    def group(self, group: Literal[0] = 0, /) -> AnyStr:
        """
        group([group1, ...]) -> str or tuple.
        Return subgroup(s) of the match by indices or names.
        For 0 returns the entire match.
        """
        ...
    @overload
    def group(self, group: str | int, /) -> AnyStr | MaybeNone:
        """
        group([group1, ...]) -> str or tuple.
        Return subgroup(s) of the match by indices or names.
        For 0 returns the entire match.
        """
        ...
    @overload
    def group(self, group1: str | int, group2: str | int, /, *groups: str | int) -> tuple[AnyStr | MaybeNone, ...]:
        """
        group([group1, ...]) -> str or tuple.
        Return subgroup(s) of the match by indices or names.
        For 0 returns the entire match.
        """
        ...
    # Each item of groups()'s return tuple is either "AnyStr" or
    # "AnyStr | None", depending on the pattern.
    @overload
    def groups(self) -> tuple[AnyStr | MaybeNone, ...]:
        """
        Return a tuple containing all the subgroups of the match, from 1.

        default
          Is used for groups that did not participate in the match.
        """
        ...
    @overload
    def groups(self, default: _T) -> tuple[AnyStr | _T, ...]:
        """
        Return a tuple containing all the subgroups of the match, from 1.

        default
          Is used for groups that did not participate in the match.
        """
        ...
    # Each value in groupdict()'s return dict is either "AnyStr" or
    # "AnyStr | None", depending on the pattern.
    @overload
    def groupdict(self) -> dict[str, AnyStr | MaybeNone]:
        """
        Return a dictionary containing all the named subgroups of the match, keyed by the subgroup name.

        default
          Is used for groups that did not participate in the match.
        """
        ...
    @overload
    def groupdict(self, default: _T) -> dict[str, AnyStr | _T]:
        """
        Return a dictionary containing all the named subgroups of the match, keyed by the subgroup name.

        default
          Is used for groups that did not participate in the match.
        """
        ...
    def start(self, group: int | str = 0, /) -> int:
        """Return index of the start of the substring matched by group."""
        ...
    def end(self, group: int | str = 0, /) -> int:
        """Return index of the end of the substring matched by group."""
        ...
    def span(self, group: int | str = 0, /) -> tuple[int, int]:
        """For match object m, return the 2-tuple (m.start(group), m.end(group))."""
        ...
    @property
    def regs(self) -> tuple[tuple[int, int], ...]: ...  # undocumented
    # __getitem__() returns "AnyStr" or "AnyStr | None", depending on the pattern.
    @overload
    def __getitem__(self, key: Literal[0], /) -> AnyStr:
        """Return self[key]."""
        ...
    @overload
    def __getitem__(self, key: int | str, /) -> AnyStr | MaybeNone:
        """Return self[key]."""
        ...
    def __copy__(self) -> Match[AnyStr]: ...
    def __deepcopy__(self, memo: Any, /) -> Match[AnyStr]: ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias:
        """See PEP 585"""
        ...

@final
class Pattern(Generic[AnyStr]):
    @property
    def flags(self) -> int:
        """The regex matching flags."""
        ...
    @property
    def groupindex(self) -> Mapping[str, int]:
        """A dictionary mapping group names to group numbers."""
        ...
    @property
    def groups(self) -> int:
        """The number of capturing groups in the pattern."""
        ...
    @property
    def pattern(self) -> AnyStr:
        """The pattern string from which the RE object was compiled."""
        ...
    @overload
    def search(self: Pattern[str], string: str, pos: int = 0, endpos: int = sys.maxsize) -> Match[str] | None:
        """
        Scan through string looking for a match, and return a corresponding match object instance.

        Return None if no position in the string matches.
        """
        ...
    @overload
    def search(self: Pattern[bytes], string: ReadableBuffer, pos: int = 0, endpos: int = sys.maxsize) -> Match[bytes] | None:
        """
        Scan through string looking for a match, and return a corresponding match object instance.

        Return None if no position in the string matches.
        """
        ...
    @overload
    def search(self, string: AnyStr, pos: int = 0, endpos: int = sys.maxsize) -> Match[AnyStr] | None:
        """
        Scan through string looking for a match, and return a corresponding match object instance.

        Return None if no position in the string matches.
        """
        ...
    @overload
    def match(self: Pattern[str], string: str, pos: int = 0, endpos: int = sys.maxsize) -> Match[str] | None:
        """Matches zero or more characters at the beginning of the string."""
        ...
    @overload
    def match(self: Pattern[bytes], string: ReadableBuffer, pos: int = 0, endpos: int = sys.maxsize) -> Match[bytes] | None:
        """Matches zero or more characters at the beginning of the string."""
        ...
    @overload
    def match(self, string: AnyStr, pos: int = 0, endpos: int = sys.maxsize) -> Match[AnyStr] | None:
        """Matches zero or more characters at the beginning of the string."""
        ...
    @overload
    def fullmatch(self: Pattern[str], string: str, pos: int = 0, endpos: int = sys.maxsize) -> Match[str] | None:
        """Matches against all of the string."""
        ...
    @overload
    def fullmatch(
        self: Pattern[bytes], string: ReadableBuffer, pos: int = 0, endpos: int = sys.maxsize
    ) -> Match[bytes] | None:
        """Matches against all of the string."""
        ...
    @overload
    def fullmatch(self, string: AnyStr, pos: int = 0, endpos: int = sys.maxsize) -> Match[AnyStr] | None:
        """Matches against all of the string."""
        ...
    @overload
    def split(self: Pattern[str], string: str, maxsplit: int = 0) -> list[str | MaybeNone]:
        """Split string by the occurrences of pattern."""
        ...
    @overload
    def split(self: Pattern[bytes], string: ReadableBuffer, maxsplit: int = 0) -> list[bytes | MaybeNone]:
        """Split string by the occurrences of pattern."""
        ...
    @overload
    def split(self, string: AnyStr, maxsplit: int = 0) -> list[AnyStr | MaybeNone]:
        """Split string by the occurrences of pattern."""
        ...
    # return type depends on the number of groups in the pattern
    @overload
    def findall(self: Pattern[str], string: str, pos: int = 0, endpos: int = sys.maxsize) -> list[Any]:
        """Return a list of all non-overlapping matches of pattern in string."""
        ...
    @overload
    def findall(self: Pattern[bytes], string: ReadableBuffer, pos: int = 0, endpos: int = sys.maxsize) -> list[Any]:
        """Return a list of all non-overlapping matches of pattern in string."""
        ...
    @overload
    def findall(self, string: AnyStr, pos: int = 0, endpos: int = sys.maxsize) -> list[AnyStr]:
        """Return a list of all non-overlapping matches of pattern in string."""
        ...
    @overload
    def finditer(self: Pattern[str], string: str, pos: int = 0, endpos: int = sys.maxsize) -> Iterator[Match[str]]:
        """
        Return an iterator over all non-overlapping matches for the RE pattern in string.

        For each match, the iterator returns a match object.
        """
        ...
    @overload
    def finditer(
        self: Pattern[bytes], string: ReadableBuffer, pos: int = 0, endpos: int = sys.maxsize
    ) -> Iterator[Match[bytes]]:
        """
        Return an iterator over all non-overlapping matches for the RE pattern in string.

        For each match, the iterator returns a match object.
        """
        ...
    @overload
    def finditer(self, string: AnyStr, pos: int = 0, endpos: int = sys.maxsize) -> Iterator[Match[AnyStr]]:
        """
        Return an iterator over all non-overlapping matches for the RE pattern in string.

        For each match, the iterator returns a match object.
        """
        ...
    @overload
    def sub(self: Pattern[str], repl: str | Callable[[Match[str]], str], string: str, count: int = 0) -> str:
        """Return the string obtained by replacing the leftmost non-overlapping occurrences of pattern in string by the replacement repl."""
        ...
    @overload
    def sub(
        self: Pattern[bytes],
        repl: ReadableBuffer | Callable[[Match[bytes]], ReadableBuffer],
        string: ReadableBuffer,
        count: int = 0,
    ) -> bytes:
        """Return the string obtained by replacing the leftmost non-overlapping occurrences of pattern in string by the replacement repl."""
        ...
    @overload
    def sub(self, repl: AnyStr | Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = 0) -> AnyStr:
        """Return the string obtained by replacing the leftmost non-overlapping occurrences of pattern in string by the replacement repl."""
        ...
    @overload
    def subn(self: Pattern[str], repl: str | Callable[[Match[str]], str], string: str, count: int = 0) -> tuple[str, int]:
        """Return the tuple (new_string, number_of_subs_made) found by replacing the leftmost non-overlapping occurrences of pattern with the replacement repl."""
        ...
    @overload
    def subn(
        self: Pattern[bytes],
        repl: ReadableBuffer | Callable[[Match[bytes]], ReadableBuffer],
        string: ReadableBuffer,
        count: int = 0,
    ) -> tuple[bytes, int]:
        """Return the tuple (new_string, number_of_subs_made) found by replacing the leftmost non-overlapping occurrences of pattern with the replacement repl."""
        ...
    @overload
    def subn(self, repl: AnyStr | Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = 0) -> tuple[AnyStr, int]:
        """Return the tuple (new_string, number_of_subs_made) found by replacing the leftmost non-overlapping occurrences of pattern with the replacement repl."""
        ...
    def __copy__(self) -> Pattern[AnyStr]: ...
    def __deepcopy__(self, memo: Any, /) -> Pattern[AnyStr]: ...
    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        ...
    def __hash__(self) -> int:
        """Return hash(self)."""
        ...
    def __class_getitem__(cls, item: Any, /) -> GenericAlias:
        """See PEP 585"""
        ...

# ----- re variables and constants -----

class RegexFlag(enum.IntFlag):
    A = sre_compile.SRE_FLAG_ASCII
    ASCII = A
    DEBUG = sre_compile.SRE_FLAG_DEBUG
    I = sre_compile.SRE_FLAG_IGNORECASE
    IGNORECASE = I
    L = sre_compile.SRE_FLAG_LOCALE
    LOCALE = L
    M = sre_compile.SRE_FLAG_MULTILINE
    MULTILINE = M
    S = sre_compile.SRE_FLAG_DOTALL
    DOTALL = S
    X = sre_compile.SRE_FLAG_VERBOSE
    VERBOSE = X
    U = sre_compile.SRE_FLAG_UNICODE
    UNICODE = U
    if sys.version_info < (3, 13):
        T = sre_compile.SRE_FLAG_TEMPLATE
        TEMPLATE = T
    if sys.version_info >= (3, 11):
        NOFLAG = 0

A: Final = RegexFlag.A
ASCII: Final = RegexFlag.ASCII
DEBUG: Final = RegexFlag.DEBUG
I: Final = RegexFlag.I
IGNORECASE: Final = RegexFlag.IGNORECASE
L: Final = RegexFlag.L
LOCALE: Final = RegexFlag.LOCALE
M: Final = RegexFlag.M
MULTILINE: Final = RegexFlag.MULTILINE
S: Final = RegexFlag.S
DOTALL: Final = RegexFlag.DOTALL
X: Final = RegexFlag.X
VERBOSE: Final = RegexFlag.VERBOSE
U: Final = RegexFlag.U
UNICODE: Final = RegexFlag.UNICODE
if sys.version_info < (3, 13):
    T: Final = RegexFlag.T
    TEMPLATE: Final = RegexFlag.TEMPLATE
if sys.version_info >= (3, 11):
    NOFLAG: Final = RegexFlag.NOFLAG
_FlagsType: TypeAlias = int | RegexFlag

# Type-wise the compile() overloads are unnecessary, they could also be modeled using
# unions in the parameter types. However mypy has a bug regarding TypeVar
# constraints (https://github.com/python/mypy/issues/11880),
# which limits us here because AnyStr is a constrained TypeVar.

# pattern arguments do *not* accept arbitrary buffers such as bytearray,
# because the pattern must be hashable.
@overload
def compile(pattern: AnyStr, flags: _FlagsType = 0) -> Pattern[AnyStr]: ...
@overload
def compile(pattern: Pattern[AnyStr], flags: _FlagsType = 0) -> Pattern[AnyStr]: ...
@overload
def search(pattern: str | Pattern[str], string: str, flags: _FlagsType = 0) -> Match[str] | None: ...
@overload
def search(pattern: bytes | Pattern[bytes], string: ReadableBuffer, flags: _FlagsType = 0) -> Match[bytes] | None: ...
@overload
def match(pattern: str | Pattern[str], string: str, flags: _FlagsType = 0) -> Match[str] | None: ...
@overload
def match(pattern: bytes | Pattern[bytes], string: ReadableBuffer, flags: _FlagsType = 0) -> Match[bytes] | None: ...
@overload
def fullmatch(pattern: str | Pattern[str], string: str, flags: _FlagsType = 0) -> Match[str] | None: ...
@overload
def fullmatch(pattern: bytes | Pattern[bytes], string: ReadableBuffer, flags: _FlagsType = 0) -> Match[bytes] | None: ...
@overload
def split(pattern: str | Pattern[str], string: str, maxsplit: int = 0, flags: _FlagsType = 0) -> list[str | MaybeNone]: ...
@overload
def split(
    pattern: bytes | Pattern[bytes], string: ReadableBuffer, maxsplit: int = 0, flags: _FlagsType = 0
) -> list[bytes | MaybeNone]: ...
@overload
def findall(pattern: str | Pattern[str], string: str, flags: _FlagsType = 0) -> list[Any]: ...
@overload
def findall(pattern: bytes | Pattern[bytes], string: ReadableBuffer, flags: _FlagsType = 0) -> list[Any]: ...
@overload
def finditer(pattern: str | Pattern[str], string: str, flags: _FlagsType = 0) -> Iterator[Match[str]]: ...
@overload
def finditer(pattern: bytes | Pattern[bytes], string: ReadableBuffer, flags: _FlagsType = 0) -> Iterator[Match[bytes]]: ...
@overload
def sub(
    pattern: str | Pattern[str], repl: str | Callable[[Match[str]], str], string: str, count: int = 0, flags: _FlagsType = 0
) -> str: ...
@overload
def sub(
    pattern: bytes | Pattern[bytes],
    repl: ReadableBuffer | Callable[[Match[bytes]], ReadableBuffer],
    string: ReadableBuffer,
    count: int = 0,
    flags: _FlagsType = 0,
) -> bytes: ...
@overload
def subn(
    pattern: str | Pattern[str], repl: str | Callable[[Match[str]], str], string: str, count: int = 0, flags: _FlagsType = 0
) -> tuple[str, int]: ...
@overload
def subn(
    pattern: bytes | Pattern[bytes],
    repl: ReadableBuffer | Callable[[Match[bytes]], ReadableBuffer],
    string: ReadableBuffer,
    count: int = 0,
    flags: _FlagsType = 0,
) -> tuple[bytes, int]: ...
def escape(pattern: AnyStr) -> AnyStr: ...
def purge() -> None: ...

if sys.version_info < (3, 13):
    if sys.version_info >= (3, 11):
        @deprecated("Deprecated since Python 3.11; removed in Python 3.13. Use `re.compile()` instead.")
        def template(pattern: AnyStr | Pattern[AnyStr], flags: _FlagsType = 0) -> Pattern[AnyStr]: ...  # undocumented
    else:
        def template(pattern: AnyStr | Pattern[AnyStr], flags: _FlagsType = 0) -> Pattern[AnyStr]: ...  # undocumented



import collections

# === OrderedDict membership (in operator) ===
d = collections.OrderedDict()
d[1] = 'a'
d[2] = 'b'

# Basic key membership
assert 1 in d, 'ordereddict: existing key should be found with in'
assert 2 in d, 'ordereddict: second key should be found with in'
assert 3 not in d, 'ordereddict: missing key should not be found with in'

# String keys
d2 = collections.OrderedDict([('x', 10), ('y', 20)])
assert 'x' in d2, 'ordereddict: string key membership'
assert 'z' not in d2, 'ordereddict: missing string key not in'

# Empty OrderedDict
d3 = collections.OrderedDict()
assert 'anything' not in d3, 'ordereddict: empty dict contains nothing'

# === move_to_end preserves membership ===
d.move_to_end(1)
assert 1 in d, 'ordereddict: key still present after move_to_end'
assert list(d.keys()) == [2, 1], 'ordereddict: move_to_end changes order'

# move_to_end with last=False
d.move_to_end(1, last=False)
assert list(d.keys()) == [1, 2], 'ordereddict: move_to_end(last=False) moves to front'

# === popitem removes from membership ===
d4 = collections.OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')])
item = d4.popitem()
assert item == (3, 'c'), 'ordereddict: popitem() pops last by default'
assert 3 not in d4, 'ordereddict: popped key no longer in dict'
assert 1 in d4, 'ordereddict: remaining key still in dict'

item2 = d4.popitem(last=False)
assert item2 == (1, 'a'), 'ordereddict: popitem(last=False) pops first'
assert 1 not in d4, 'ordereddict: popped first key no longer in dict'



# Security: infinite recursion must raise RecursionError, not crash the host.
# RecursionError must be catchable (CPython parity).
# After catching RecursionError, the stack must recover for normal use.
#
# NOTE: Keep the number of actual recursion-to-limit tests small — each one
# consumes significant stack in the CPython test runner thread.

# === Direct infinite recursion raises RecursionError ===
def infinite():
    return infinite()

caught = False
try:
    infinite()
except RecursionError:
    caught = True
assert caught, 'infinite recursion must raise RecursionError'

# === RecursionError is catchable with message ===
msg = ''
try:
    infinite()
except RecursionError as e:
    msg = str(e)
assert msg == 'maximum recursion depth exceeded', f'RecursionError message wrong: {msg}'

# === Stack recovers after RecursionError ===
def safe(n):
    if n <= 0:
        return 0
    return safe(n - 1)

try:
    infinite()
except RecursionError:
    pass
assert safe(10) == 0, 'normal recursion works after catching RecursionError'

# === Infinite mutual recursion ===
def ping():
    return pong()

def pong():
    return ping()

caught = False
try:
    ping()
except RecursionError:
    caught = True
assert caught, 'mutual infinite recursion must raise RecursionError'

# === Recursion in __init__ ===
class InitRecurse:
    def __init__(self):
        self.child = InitRecurse()

caught = False
try:
    InitRecurse()
except RecursionError:
    caught = True
assert caught, '__init__ infinite recursion must raise RecursionError'

# === RecursionError caught inside the recursive function itself ===
def self_catching():
    try:
        return self_catching()
    except RecursionError:
        return 'caught'

result = self_catching()
assert result == 'caught', f'self-catching recursion should return caught, got {result}'

# === Normal code works after all the recursion errors ===
assert safe(20) == 0, 'stack fully recovered after multiple RecursionErrors'



import typing_extensions

# === Public API surface ===
expected_public_names = [
    'AbstractSet',
    'Annotated',
    'Any',
    'AnyStr',
    'AsyncContextManager',
    'AsyncGenerator',
    'AsyncIterable',
    'AsyncIterator',
    'Awaitable',
    'BinaryIO',
    'Buffer',
    'Callable',
    'CapsuleType',
    'ChainMap',
    'ClassVar',
    'Collection',
    'Concatenate',
    'Container',
    'ContextManager',
    'Coroutine',
    'Counter',
    'DefaultDict',
    'Deque',
    'Dict',
    'Doc',
    'Final',
    'Format',
    'ForwardRef',
    'FrozenSet',
    'Generator',
    'Generic',
    'GenericMeta',
    'Hashable',
    'IO',
    'IntVar',
    'ItemsView',
    'Iterable',
    'Iterator',
    'KT',
    'KeysView',
    'List',
    'Literal',
    'LiteralString',
    'Mapping',
    'MappingView',
    'Match',
    'MutableMapping',
    'MutableSequence',
    'MutableSet',
    'NamedTuple',
    'Never',
    'NewType',
    'NoDefault',
    'NoExtraItems',
    'NoReturn',
    'NotRequired',
    'Optional',
    'OrderedDict',
    'PEP_560',
    'ParamSpec',
    'ParamSpecArgs',
    'ParamSpecKwargs',
    'Pattern',
    'Protocol',
    'ReadOnly',
    'Reader',
    'Required',
    'Reversible',
    'Self',
    'Sentinel',
    'Sequence',
    'Set',
    'Sized',
    'SupportsAbs',
    'SupportsBytes',
    'SupportsComplex',
    'SupportsFloat',
    'SupportsIndex',
    'SupportsInt',
    'SupportsRound',
    'T',
    'TYPE_CHECKING',
    'T_co',
    'T_contra',
    'Text',
    'TextIO',
    'Tuple',
    'Type',
    'TypeAlias',
    'TypeAliasType',
    'TypeForm',
    'TypeGuard',
    'TypeIs',
    'TypeVar',
    'TypeVarTuple',
    'TypedDict',
    'Union',
    'Unpack',
    'VT',
    'ValuesView',
    'Writer',
    'abc',
    'annotationlib',
    'assert_never',
    'assert_type',
    'builtins',
    'cast',
    'clear_overloads',
    'collections',
    'contextlib',
    'dataclass_transform',
    'deprecated',
    'disjoint_base',
    'enum',
    'evaluate_forward_ref',
    'final',
    'functools',
    'get_annotations',
    'get_args',
    'get_origin',
    'get_original_bases',
    'get_overloads',
    'get_protocol_members',
    'get_type_hints',
    'inspect',
    'io',
    'is_protocol',
    'is_typeddict',
    'keyword',
    'no_type_check',
    'no_type_check_decorator',
    'operator',
    'overload',
    'override',
    'reveal_type',
    'runtime',
    'runtime_checkable',
    'sys',
    'type_repr',
    'typing',
    'warnings',
]

missing_names = [name for name in expected_public_names if not hasattr(typing_extensions, name)]
assert missing_names == [], f'missing typing_extensions names: {missing_names}'

dir_names = dir(typing_extensions)
for name in expected_public_names:
    assert name in dir_names, f'{name} should appear in dir(typing_extensions)'

# === Basic invariants ===
assert typing_extensions.PEP_560 == True, 'PEP_560 should be True'
assert typing_extensions.runtime is typing_extensions.runtime_checkable, 'runtime should alias runtime_checkable'
assert typing_extensions.type_repr(int) == 'int', 'type_repr(int) should be int'
assert typing_extensions.type_repr(str) == 'str', 'type_repr(str) should be str'
assert typing_extensions.NoExtraItems is not None, 'NoExtraItems should exist'
assert typing_extensions.Format.VALUE == 1, 'Format.VALUE should be 1'
assert typing_extensions.Format.STRING == 4, 'Format.STRING should be 4'

# === Module-valued exports ===
for module_name in [
    'abc',
    'annotationlib',
    'builtins',
    'collections',
    'contextlib',
    'enum',
    'functools',
    'inspect',
    'io',
    'keyword',
    'operator',
    'sys',
    'typing',
    'warnings',
]:
    module_value = getattr(typing_extensions, module_name)
    assert module_value is not None, f'{module_name} should be present'

# === TypeVar-like exports ===
for symbol_name in ['T', 'KT', 'VT', 'T_co', 'T_contra']:
    symbol_value = getattr(typing_extensions, symbol_name)
    assert hasattr(symbol_value, '__name__'), f'{symbol_name} should have __name__'
    assert symbol_value.__name__ == symbol_name, f'{symbol_name} should preserve symbol name'

int_var = typing_extensions.IntVar('MyInt')
assert hasattr(int_var, '__name__'), 'IntVar result should have __name__'
assert int_var.__name__ == 'MyInt', 'IntVar should preserve the provided name'
assert isinstance(int_var.__constraints__, tuple), 'IntVar __constraints__ should be tuple'

# === Runtime helper functions ===
assert typing_extensions.cast(int, 'x') == 'x', 'cast should return value unchanged'
sample = ['a', 'b']
assert typing_extensions.assert_type(sample, typing_extensions.Sequence) is sample, 'assert_type should be identity'
assert typing_extensions.reveal_type(sample) is sample, 'reveal_type should return original value'
assert typing_extensions.get_origin(int) is None, 'get_origin(int) should be None'
assert typing_extensions.get_args(int) == (), 'get_args(int) should be empty tuple'

annotation_result = None


def annotated_func(x: int) -> str:
    return str(x)


annotation_result = typing_extensions.get_annotations(annotated_func)
assert annotation_result == {'x': int, 'return': str}, 'get_annotations should return __annotations__ mapping'

forward_ref = typing_extensions.ForwardRef('int')
assert typing_extensions.evaluate_forward_ref(forward_ref) is int, 'evaluate_forward_ref should resolve builtins'

assert_never_message = None
try:
    typing_extensions.assert_never(123)
except AssertionError as exc:
    assert_never_message = exc.args[0]
assert (
    assert_never_message == 'Expected code to be unreachable, but got: 123'
), 'assert_never should raise AssertionError with CPython-compatible text'

# === Decorator helpers ===


@typing_extensions.disjoint_base
class DisjointBaseType:
    pass


assert DisjointBaseType.__name__ == 'DisjointBaseType', 'disjoint_base should return class unchanged'


@typing_extensions.deprecated('legacy API')
def legacy_fn(x):
    return x + 1


assert callable(legacy_fn), 'deprecated decorator should return a callable'

# === Class-ish helper constructors ===
doc = typing_extensions.Doc('some documentation')
assert doc.documentation == 'some documentation', 'Doc should expose documentation attribute'

sentinel_a = typing_extensions.Sentinel('TOKEN')
sentinel_b = typing_extensions.Sentinel('TOKEN')
assert sentinel_a is not sentinel_b, 'Sentinel should produce unique instances'
assert typing_extensions.TypeForm(int) is int, 'TypeForm should be runtime identity'

# === TypedDict and Protocol helpers ===


class Proto(typing_extensions.Protocol):
    def run(self):
        pass


assert typing_extensions.is_protocol(Proto) == True, 'is_protocol should recognize protocol classes'


@typing_extensions.runtime_checkable
class RuntimeProto(typing_extensions.Protocol):
    def run(self):
        pass


class Runner:
    def run(self):
        return 'ok'


assert isinstance(Runner(), RuntimeProto) == True, 'runtime_checkable protocols should support isinstance'


class TD(typing_extensions.TypedDict):
    value: int


assert typing_extensions.is_typeddict(TD) == True, 'is_typeddict should recognize TypedDict classes'
td_instance = TD(value=10)
assert td_instance == {'value': 10}, 'TypedDict class should construct dict values'

# === Misc helpers ===


class PlainClass:
    pass


original_bases = typing_extensions.get_original_bases(PlainClass)
assert isinstance(original_bases, tuple), 'get_original_bases should return a tuple'



import contextlib

# === function existence checks ===
assert contextlib.nullcontext is not None, 'nullcontext should exist'
assert contextlib.closing is not None, 'closing should exist'
assert contextlib.aclosing is not None, 'aclosing should exist'
assert contextlib.suppress is not None, 'suppress should exist'
assert contextlib.redirect_stdout is not None, 'redirect_stdout should exist'
assert contextlib.redirect_stderr is not None, 'redirect_stderr should exist'
assert contextlib.contextmanager is not None, 'contextmanager should exist'
assert contextlib.ExitStack is not None, 'ExitStack should exist'
assert contextlib.AsyncExitStack is not None, 'AsyncExitStack should exist'

# === closing returns something ===
obj = [1, 2, 3]
result = contextlib.closing(obj)
assert result is not None, 'closing should return something'

async_result = contextlib.aclosing(obj)
assert async_result is not None, 'aclosing should return a value'

# === ExitStack basic API ===
stack = contextlib.ExitStack()
assert stack is not None, 'ExitStack() returns an object'
assert stack.push(lambda *exc: None) is not None, 'push returns callback/context manager'
assert stack.callback(lambda: None) is not None, 'callback returns callback function'
assert stack.close() is None, 'close returns None'

with contextlib.ExitStack() as entered:
    assert entered is not None, '__enter__ should return a value'

async_stack = contextlib.AsyncExitStack()
assert async_stack is not None, 'AsyncExitStack() returns an object'

# === suppress exception matching ===
continued_after_suppress = False
with contextlib.suppress((ValueError, TypeError)):
    raise TypeError('suppressed')
continued_after_suppress = True
assert continued_after_suppress, 'suppress should handle tuple exception classinfo'



','.join([1, 2])
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__join_non_string.py", line 1, in <module>
    ','.join([1, 2])
    ~~~~~~~~~~~~~~~~
TypeError: sequence item 0: expected str instance, int found
"""



# call-external
# Test external call with exception variable

exc_with_ext = None
try:
    raise ValueError('test')
except ValueError as e:
    prefix = concat_strings('caught: ', repr(e))
    exc_with_ext = prefix
assert exc_with_ext == "caught: ValueError('test')", 'ext call with exception variable'



import functools

# === reduce with lambda ===
result = functools.reduce(lambda a, b: a + b, [1, 2, 3, 4])
assert result == 10, 'reduce sum'

result = functools.reduce(lambda a, b: a * b, [1, 2, 3, 4])
assert result == 24, 'reduce product'

# === reduce with initial ===
result = functools.reduce(lambda a, b: a + b, [1, 2, 3], 10)
assert result == 16, 'reduce with initial'

# === reduce single element ===
result = functools.reduce(lambda a, b: a + b, [42])
assert result == 42, 'reduce single'

# === from import ===
from functools import reduce

assert reduce(lambda a, b: a + b, [1, 2]) == 3, 'from import reduce'



def recurse():
    recurse()


recurse()
"""
TRACEBACK:
Traceback (most recent call last):
  File "traceback__recursion_error.py", line 5, in <module>
    recurse()
    ~~~~~~~~~
  File "traceback__recursion_error.py", line 2, in recurse
    recurse()
    ~~~~~~~~~
  File "traceback__recursion_error.py", line 2, in recurse
    recurse()
    ~~~~~~~~~
  File "traceback__recursion_error.py", line 2, in recurse
    recurse()
    ~~~~~~~~~
  [Previous line repeated 47 more times]
RecursionError: maximum recursion depth exceeded
"""



# === Dict literals ===
assert {} == {}, 'empty literal'
assert {'a': 1} == {'a': 1}, 'single item literal'
assert {'a': 1, 'b': 2} == {'a': 1, 'b': 2}, 'multiple items literal'
assert {1: 'a', 2: 'b'} == {1: 'a', 2: 'b'}, 'int keys literal'

# === Dict length ===
assert len({}) == 0, 'len empty'
assert len({'a': 1, 'b': 2, 'c': 3}) == 3, 'len multiple'

# === Dict equality ===
assert ({'a': 1, 'b': 2} == {'b': 2, 'a': 1}) == True, 'equality true (order independent)'
assert ({'a': 1} == {'a': 2}) == False, 'equality false'

# === Dict subscript get ===
d = {'name': 'Alice', 'age': 30}
assert d['name'] == 'Alice', 'subscript get str key'
assert d['age'] == 30, 'subscript get value'

d = {1: 'one', 2: 'two'}
assert d[1] == 'one', 'subscript get int key'

# === Dict subscript set ===
d = {'a': 1}
d['b'] = 2
assert d == {'a': 1, 'b': 2}, 'subscript set new key'

d = {'a': 1}
d['a'] = 99
assert d == {'a': 99}, 'subscript set existing key'

# === Dict.get() method ===
d = {'a': 1, 'b': 2}
assert d.get('a') == 1, 'get existing'
assert d.get('missing') is None, 'get missing returns None'
assert d.get('missing', 'default') == 'default', 'get missing with default'

# === Dict.pop() method ===
d = {'a': 1, 'b': 2}
assert d.pop('a') == 1, 'pop existing'
assert d == {'b': 2}, 'pop removes key'

d = {'a': 1}
assert d.pop('missing', 'default') == 'default', 'pop missing with default'

# === Dict with tuple key ===
d = {(1, 2): 'value'}
assert d[(1, 2)] == 'value', 'tuple key'

# === Dict repr ===
assert repr({}) == '{}', 'empty repr'
assert repr({'a': 1}) == "{'a': 1}", 'repr with items'

# === Dict self-reference ===
d = {}
d['self'] = d
assert d['self'] is d, 'getitem self'

d = {}
assert d.get('missing', d) is d, 'get default same dict'



x = len
x.nonexistent
"""
TRACEBACK:
Traceback (most recent call last):
  File "type__builtin_attr_error.py", line 2, in <module>
    x.nonexistent
AttributeError: 'builtin_function_or_method' object has no attribute 'nonexistent'
"""



# run-async
await 123  # pyright: ignore
"""
TRACEBACK:
Traceback (most recent call last):
  File "async__not_awaitable.py", line 2, in <module>
    await 123  # pyright: ignore
    ~~~~~~~~~
TypeError: 'int' object can't be awaited
"""
